# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Multimodal data features types and query mechanisms."""

from typing import Union
from functools import partial

from . import (
    connectivity,
    tabular,
    image,
    dataset,
)

from ..commons import logger

from .feature import Feature
from ..retrieval import cache
from ..commons import siibra_tqdm

get = Feature._match

TYPES = Feature._get_subclasses()  # Feature types that can be used to query for features


def __dir__():
    return list(Feature._CATEGORIZED.keys()) + ["get", "TYPES", "render_ascii_tree"]


def __getattr__(attr: str):
    if attr in Feature._CATEGORIZED:
        return Feature._CATEGORIZED[attr]
    else:
        hint = ""
        if isinstance(attr, str):
            import difflib
            closest = difflib.get_close_matches(attr, list(__dir__()), n=3)
            if len(closest) > 0:
                hint = f"Did you mean {' or '.join(closest)}?"
        raise AttributeError(f"No such attribute: {__name__}.{attr} " + hint)


@cache.Warmup.register_warmup_fn()
def _warm_feature_cache_instances():
    """Preload preconfigured multimodal data features."""
    for ftype in TYPES.values():
        _ = ftype._get_instances()


@cache.Warmup.register_warmup_fn(cache.WarmupLevel.DATA, is_factory=True)
def _warm_feature_cache_data():
    return_callables = []
    for ftype in TYPES.values():
        instances = ftype._get_instances()

        # the instances *must* be cleared, or it will impede the garbage collection, and results in memleak
        ftype._clean_instances()
        tally = siibra_tqdm(desc=f"Warming data {ftype.__name__}", total=len(instances))
        for f in instances:
            def get_data(arg):
                tally = arg.pop("tally")
                feature = arg.pop("feature")
                # TODO
                # the try catch is as a result of https://github.com/FZJ-INM1-BDA/siibra-python/issues/509
                # sometimes f.data can fail
                try:
                    _ = feature.data
                except Exception as e:
                    logger.warn(f"Feature {feature.name} warmup failed: {str(e)}")
                finally:
                    tally.update(1)
            # append dictionary, so that popping the dictionary will mark the feature to be garbage collected
            return_callables.append(partial(get_data, {"feature": f, "tally": tally}))
    return return_callables


def render_ascii_tree(class_or_classname: Union[type, str]):
    """
    Print the ascii hierarchy representation of a feature type.

    Parameters
    ----------
    class_or_classname: type, str
        Any Feature class or string of the feature type name
    """
    from anytree.importer import DictImporter
    from anytree import RenderTree
    Cls = TYPES[class_or_classname] if isinstance(class_or_classname, str) else class_or_classname
    assert issubclass(Cls, Feature)

    def create_treenode(feature_type):
        return {
            'name': feature_type.__name__,
            'children': [
                create_treenode(c)
                for c in feature_type.__subclasses__()
            ]
        }
    D = create_treenode(Cls)
    importer = DictImporter()
    tree = importer.import_(D)
    print("\n".join(
        "%s%s" % (pre, node.name)
        for pre, _, node in RenderTree(tree)
    ))



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handles reading and preparing gii files."""

from typing import Union, Dict, TYPE_CHECKING

import numpy as np

from . import provider as _provider
from ...retrieval.requests import HttpRequest, ZipfileRequest

if TYPE_CHECKING:
    from ...locations import boundingbox as _boundingbox


class FreesurferAnnot(_provider.VolumeProvider, srctype="freesurfer-annot"):
    def __init__(self, url: Union[str, dict]):
        self._init_url = url
        if isinstance(url, str):  # single mesh labelling
            self._loaders = {None: HttpRequest(url)}
        elif isinstance(url, dict):   # named label fragments
            self._loaders = {lbl: HttpRequest(u) for lbl, u in url.items()}
        else:
            raise NotImplementedError(f"Urls for {self.__class__.__name__} are expected to be of type str.")

    def fetch(self, fragment: str = None, label: int = None, **kwargs):
        """Returns a 1D numpy array of label indices."""
        vertex_labels = []
        if fragment is None:
            matched_frags = list(self._loaders.keys())
        else:
            matched_frags = [frg for frg in self._loaders.keys() if fragment.lower() in frg.lower()]
            if len(matched_frags) != 1:
                raise ValueError(
                    f"Requested fragment '{fragment}' could not be matched uniquely "
                    f"to [{', '.join(self._loaders)}]"
                )
        for frag in matched_frags:
            frag_labels, *_ = self._loaders[frag].data
            if label is not None:  # create the mask
                selected_label = frag_labels == label
                frag_labels[selected_label] = 1
                frag_labels[~selected_label] = 0
            else:
                frag_labels[frag_labels == -1] = 0  # annot files store background as -1 while siibra uses 0
            vertex_labels.append(frag_labels)

        return {"labels": np.hstack(vertex_labels)}

    def get_boundingbox(self, clip=False, background=0.0) -> '_boundingbox.BoundingBox':
        raise NotImplementedError(
            f"Bounding box access to {self.__class__.__name__} objects not yet implemented."
        )

    @property
    def _url(self) -> Union[str, Dict[str, str]]:
        return self._init_url


class ZippedFreesurferAnnot(_provider.VolumeProvider, srctype="zip/freesurfer-annot"):
    def __init__(self, url: Union[str, dict]):
        self._init_url = url
        if isinstance(url, str):  # single mesh labelling
            self._loaders = {None: ZipfileRequest(*url.split(" "))}
        elif isinstance(url, dict):   # named label fragments
            self._loaders = {lbl: ZipfileRequest(*u.split(" ")) for lbl, u in url.items()}
        else:
            raise NotImplementedError(f"Urls for {self.__class__.__name__} are expected to be of type str.")

    def fetch(self, fragment: str = None, label: int = None, **kwargs):
        """Returns a 1D numpy array of label indices."""
        vertex_labels = []
        if fragment is None:
            matched_frags = list(self._loaders.keys())
        else:
            matched_frags = [frg for frg in self._loaders.keys() if fragment.lower() in frg.lower()]
            if len(matched_frags) != 1:
                raise ValueError(
                    f"Requested fragment '{fragment}' could not be matched uniquely "
                    f"to [{', '.join(self._loaders)}]"
                )
        for frag in matched_frags:
            frag_labels, *_ = self._loaders[frag].data
            if label is not None:  # create the mask
                selected_label = frag_labels == label
                frag_labels[selected_label] = 1
                frag_labels[~selected_label] = 0
            else:
                frag_labels[frag_labels == -1] = 0  # annot files store background as -1 while siibra uses 0
            vertex_labels.append(frag_labels)

        return {"labels": np.hstack(vertex_labels)}

    def get_boundingbox(self, clip=False, background=0.0) -> '_boundingbox.BoundingBox':
        raise NotImplementedError(
            f"Bounding box access to {self.__class__.__name__} objects not yet implemented."
        )

    @property
    def _url(self) -> Union[str, Dict[str, str]]:
        return self._init_url



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A set of coordinates on a reference space."""

from typing import List, Union, Tuple, Dict, TYPE_CHECKING
import numbers
import json

import numpy as np
try:
    from sklearn.cluster import HDBSCAN
    _HAS_HDBSCAN = True
except ImportError:
    import sklearn
    _HAS_HDBSCAN = False

from . import location, point, boundingbox as _boundingbox
from ..retrieval.requests import HttpRequest
from ..commons import logger
from ..exceptions import SpaceWarpingFailedError, EmptyPointCloudError

if TYPE_CHECKING:
    from ..core.space import Space


def from_points(points: List["point.Point"], newlabels: List[Union[int, float, tuple]] = None) -> "PointCloud":
    """
    Create a PointCloud from an iterable of Points.

    Parameters
    ----------
    points : Iterable[point.Point]
    newlabels: List[int], optional
        Use these labels instead of the original labels of the points.

    Returns
    -------
    PointCloud
    """
    if len(points) == 0:
        raise EmptyPointCloudError("Cannot create a PointCloud without any points.")

    spaces = {p.space for p in points}
    assert len(spaces) == 1, f"PointCloud can only be constructed with points from the same space.\n{spaces}"
    coords, sigmas, labels = zip(*((p.coordinate, p.sigma, p.label) for p in points))
    if all(lb is None for lb in set(labels)):
        labels = None
    return PointCloud(
        coordinates=coords,
        space=next(iter(spaces)),
        sigma_mm=sigmas,
        labels=newlabels or labels
    )


class PointCloud(location.Location):
    """A set of 3D points in the same reference space,
    defined by a list of coordinates."""

    def __init__(
        self,
        coordinates: Union[List[Tuple], np.ndarray],
        space=None,
        sigma_mm: Union[int, float, List[Union[int, float]]] = 0,
        labels: List[Union[int, float, tuple]] = None
    ):
        """
        Construct a 3D point set in the given reference space.

        Parameters
        ----------
        coordinates : array-like, Nx3
            Coordinates in mm of the given space
        space : reference space (id, name, or Space object)
            The reference space
        sigma_mm : float, or list of float
            Optional standard deviation of point locations.
        labels: list of point labels (optional)
        """
        location.Location.__init__(self, space)

        if len(coordinates) == 0:
            raise EmptyPointCloudError(f"Cannot create a {self.__class__.__name__} without any coordinates.")

        self._coordinates = coordinates
        if not isinstance(coordinates, np.ndarray):
            self._coordinates = np.array(self._coordinates).reshape((-1, 3))
        assert len(self._coordinates.shape) == 2
        assert self._coordinates.shape[1] == 3

        if isinstance(sigma_mm, numbers.Number):
            self.sigma_mm = [sigma_mm for _ in range(len(self))]
        else:
            assert len(sigma_mm) == len(self), "The number of coordinate must be equal to the number of sigmas."
            self.sigma_mm = sigma_mm

        if labels is not None:
            assert len(labels) == self._coordinates.shape[0]
        self.labels = labels

    def intersection(self, other: location.Location):
        """Return the subset of points that are inside the given mask.

        NOTE: The affine matrix of the image must be set to warp voxels
        coordinates into the reference space of this Bounding Box.
        """
        if not isinstance(other, (point.Point, PointCloud, _boundingbox.BoundingBox)):
            return other.intersection(self)

        if isinstance(other, PointCloud):
            intersecting_points = [p for p in self if p.coordinate in other.coordinates]
        elif isinstance(other, point.Point):
            return other if other in self else None
        else:
            intersecting_points = [p for p in self if p.intersects(other)]
        if len(intersecting_points) == 0:
            return None
        intersection = from_points(intersecting_points)
        return intersection[0] if len(intersection) == 1 else intersection

    @property
    def coordinates(self) -> np.ndarray:
        return self._coordinates

    @property
    def sigma(self) -> List[Union[int, float]]:
        """The list of sigmas corresponding to the points."""
        return self.sigma_mm

    @property
    def has_constant_sigma(self) -> bool:
        return len(set(self.sigma)) == 1

    def warp(self, space: Union[str, Dict, "Space"], chunksize: int = 1000):
        """Creates a new point set by warping its points to another space"""
        from ..core.space import Space

        spaceobj = space if isinstance(space, Space) else Space.get_instance(space)
        if spaceobj == self.space:
            return self
        if any(_ not in location.Location.SPACEWARP_IDS for _ in [self.space.id, spaceobj.id]):
            raise SpaceWarpingFailedError(
                f"Cannot convert coordinates between {self.space.id} and {spaceobj.id}"
            )

        src_points = self.as_list()
        tgt_points = []
        N = len(src_points)
        if N > 10e5:
            logger.info(f"Warping {N} points from {self.space.name} to {spaceobj.name} space")
        for i0 in range(0, N, chunksize):

            i1 = min(i0 + chunksize, N)
            data = json.dumps({
                "source_space": location.Location.SPACEWARP_IDS[self.space.id],
                "target_space": location.Location.SPACEWARP_IDS[spaceobj.id],
                "source_points": src_points[i0:i1]
            })
            response = HttpRequest(
                url=f"{location.Location.SPACEWARP_SERVER}/transform-points",
                post=True,
                headers={
                    "accept": "application/json",
                    "Content-Type": "application/json",
                },
                data=data,
                func=lambda b: json.loads(b.decode()),
            ).data
            tgt_points.extend(list(response["target_points"]))

        # TODO: consider using np.isnan(np.dot(arr, arr)). see https://stackoverflow.com/a/45011547
        if np.any(np.isnan(response['target_points'])):
            raise SpaceWarpingFailedError(f'Warping {str(self)} to {spaceobj.name} resulted in NaN')

        return self.__class__(coordinates=tuple(tgt_points), space=spaceobj, labels=self.labels)

    def transform(self, affine: np.ndarray, space=None):
        """Returns a new PointCloud obtained by transforming the
        coordinates of this one with the given affine matrix.

        Parameters
        ----------
        affine : numpy 4x4 ndarray
            affine matrix
        space : reference space (id, name, or Space)
            Target reference space which is reached after
            applying the transform. Note that the consistency
            of this cannot be checked and is up to the user.
        """
        return self.__class__(
            np.dot(affine, self.homogeneous.T)[:3, :].T,
            space,
            labels=self.labels
        )

    def __getitem__(self, index: int):
        if (abs(index) >= self.__len__()):
            raise IndexError(
                f"pointcloud with {self.__len__()} points "
                f"cannot be accessed with index {index}."
            )
        return point.Point(
            self.coordinates[index, :],
            space=self.space,
            sigma_mm=self.sigma_mm[index],
            label=None if self.labels is None else self.labels[index]
        )

    def __iter__(self):
        """Return an iterator over the coordinate locations."""
        return (
            point.Point(
                self.coordinates[i, :],
                space=self.space,
                sigma_mm=self.sigma_mm[i],
                label=None if self.labels is None else self.labels[i]
            )
            for i in range(len(self))
        )

    def __eq__(self, other: 'PointCloud'):
        if isinstance(other, point.Point):
            return len(self) == 1 and self[0] == other
        if not isinstance(other, PointCloud):
            return False
        return list(self) == list(other)

    def __hash__(self):
        return super().__hash__()

    def __len__(self):
        """The number of points in this PointCloud."""
        return self.coordinates.shape[0]

    def __str__(self):
        return f"Set of {len(self)} points in the {self.boundingbox}"

    @property
    def boundingbox(self):
        """
        Return the bounding box of these points, or None in the
        special case of an empty PointCloud.
        """
        if len(self.coordinates) == 0:
            return None
        coords = self.coordinates
        # TODO this needs a more precise treatment of the sigmas
        sigma_min = max(self.sigma[i] for i in coords.argmin(0))
        sigma_max = max(self.sigma[i] for i in coords.argmax(0))
        return _boundingbox.BoundingBox(
            point1=coords.min(0),
            point2=coords.max(0),
            space=self.space,
            sigma_mm=[sigma_min, sigma_max]
        )

    @property
    def centroid(self):
        return point.Point(self.coordinates.mean(0), space=self.space)

    @property
    def volume(self):
        if len(self) < 2:
            return 0
        else:
            return self.boundingbox.volume

    def as_list(self):
        """Return the point set as a list of 3D tuples."""
        return list(zip(*self.coordinates.T.tolist()))

    @property
    def homogeneous(self):
        """Access the list of 3D point as an Nx4 array of homogeneous coordinates."""
        return np.c_[self.coordinates, np.ones(len(self))]

    def find_clusters(
        self,
        min_fraction: float = 1 / 200,
        max_fraction: float = 1 / 8
    ) -> List[int]:
        """
        Find clusters using HDBSCAN (https://dl.acm.org/doi/10.1145/2733381)
        implementation of scikit-learn (https://dl.acm.org/doi/10.5555/1953048.2078195).

        Parameters
        ----------
        min_fraction: min cluster size as a fraction of total points in the PointCloud
        max_fraction: max cluster size as a fraction of total points in the PointCloud

        Returns
        -------
        List[int]
            Returns the cluster labels found by skilearn.cluster.HDBSCAN.

            Note
            ----
            Replaces the labels of the PointCloud instance with these labels.

        Raises
        ------
        RuntimeError
            If a sklearn version without HDBSCAN is installed.
        """
        if not _HAS_HDBSCAN:
            raise RuntimeError(
                f"HDBSCAN is not available with your version {sklearn.__version__} "
                "of sckit-learn. `PointCloud.find_clusters()` will not be available."
            )
        points = np.array(self.as_list())
        N = points.shape[0]
        clustering = HDBSCAN(
            min_cluster_size=int(N * min_fraction),
            max_cluster_size=int(N * max_fraction),
        )
        if self.labels is not None:
            logger.warning(
                "Existing labels of PointCloud will be overwritten with cluster labels."
            )
        self.labels = clustering.fit_predict(points)
        return self.labels

    @property
    def label_colors(self):
        """ return a color for the given label. """
        if self.labels is None:
            return None
        else:
            try:
                from matplotlib.pyplot import cm as colormaps
            except Exception:
                logger.error("Matplotlib is not available. Label colors is disabled.")
                return None
            return colormaps.rainbow(np.linspace(0, 1, max(self.labels) + 1))


class Contour(PointCloud):
    """
    A PointCloud that represents a contour line.
    The only difference is that the point order is relevant,
    and consecutive points are thought as being connected by an edge.

    In fact, PointCloud assumes order as well, but no connections between points.
    """

    def __init__(self, coordinates, space=None, sigma_mm=0, labels: list = None):
        PointCloud.__init__(self, coordinates, space, sigma_mm, labels)

    def crop(self, voi: "_boundingbox.BoundingBox") -> List["Contour"]:
        """
        Crop the contour with a volume of interest.
        Since the contour might be split from the cropping,
        returns a set of contour segments.
        """
        segments = []

        # set the contour point labels to a linear numbering
        # so we can use them after the intersection to detect splits.
        old_labels = self.labels
        self.labels = list(range(len(self)))
        cropped = self.intersection(voi)

        if cropped is not None and not isinstance(cropped, point.Point):
            assert isinstance(cropped, PointCloud)
            # Identify contour splits are by discontinuouities ("jumps")
            # of their labels, which denote positions in the original contour
            jumps = np.diff([self.labels.index(lb) for lb in cropped.labels])
            splits = [0] + list(np.where(jumps > 1)[0] + 1) + [len(cropped)]
            for i, j in zip(splits[:-1], splits[1:]):
                segments.append(
                    self.__class__(cropped.coordinates[i:j, :], space=cropped.space)
                )

        # reset labels of the input contour points.
        self.labels = old_labels

        return segments



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Maintaining and handling caching files on disk."""

import hashlib
import os
from appdirs import user_cache_dir
import tempfile
from functools import wraps
from enum import Enum
from typing import Callable, List, NamedTuple, Union
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

from filelock import FileLock as Lock

from ..commons import logger, SIIBRA_CACHEDIR, SKIP_CACHEINIT_MAINTENANCE, siibra_tqdm
from ..exceptions import WarmupRegException


def assert_folder(folder):
    # make sure the folder exists and is writable, then return it.
    # If it cannot be written, create and return
    # a temporary folder.
    try:
        if not os.path.isdir(folder):
            os.makedirs(folder)
        if not os.access(folder, os.W_OK):
            raise OSError
        return folder
    except OSError:
        # cannot write to requested directory, create a temporary one.
        tmpdir = tempfile.mkdtemp(prefix="siibra-cache-")
        logger.warning(
            f"Siibra created a temporary cache directory at {tmpdir}, as "
            f"the requested folder ({folder}) was not usable. "
            "Please consider to set the SIIBRA_CACHEDIR environment variable "
            "to a suitable directory.")
        return tmpdir


class Cache:

    _instance = None
    folder = user_cache_dir(".".join(__name__.split(".")[:-1]), "")
    SIZE_GIB = 2  # maintenance will delete old files to stay below this limit

    def __init__(self):
        raise RuntimeError(
            "Call instance() to access "
            f"{self.__class__.__name__}")

    @classmethod
    def instance(cls):
        """
        Return an instance of the siibra cache. Create folder if needed.
        """
        if cls._instance is None:
            if SIIBRA_CACHEDIR:
                cls.folder = SIIBRA_CACHEDIR
            cls.folder = assert_folder(cls.folder)
            cls._instance = cls.__new__(cls)
            if SKIP_CACHEINIT_MAINTENANCE:
                logger.debug("Will not run maintenance on cache as SKIP_CACHE_MAINTENANCE is set to True.")
            else:
                cls._instance.run_maintenance()
        return cls._instance

    def clear(self):
        import shutil

        logger.info(f"Clearing siibra cache at {self.folder}")
        shutil.rmtree(self.folder)
        self.folder = assert_folder(self.folder)

    def run_maintenance(self):
        """ Shrinks the cache by deleting oldest files first until the total size
        is below cache size (Cache.SIZE) given in GiB."""
        # build sorted list of cache files and their os attributes
        files = [os.path.join(self.folder, fname) for fname in os.listdir(self.folder)]
        sfiles = sorted([(fn, os.stat(fn)) for fn in files], key=lambda t: t[1].st_atime)

        # determine the first n files that need to be deleted to reach the accepted cache size
        size_gib = sum(t[1].st_size for t in sfiles) / 1024**3
        targetsize = size_gib
        index = 0
        for index, (fn, st) in enumerate(sfiles):
            if targetsize <= self.SIZE_GIB:
                break
            targetsize -= st.st_size / 1024**3

        if index > 0:
            logger.debug(f"Removing the {index + 1} oldest files to keep cache size below {targetsize:.2f} GiB.")
            for fn, st in sfiles[:index + 1]:
                if os.path.isdir(fn):
                    import shutil
                    size = sum(os.path.getsize(f) for f in os.listdir(fn) if os.path.isfile(f))
                    shutil.rmtree(fn)
                else:
                    size = st.st_size
                    os.remove(fn)
                size_gib -= size / 1024**3

    @property
    def size(self):
        """ Return size of the cache in GiB. """
        return sum(os.path.getsize(fn) for fn in self) / 1024**3

    def __iter__(self):
        """ Iterate all element names in the cache directory. """
        return (os.path.join(self.folder, f) for f in os.listdir(self.folder))

    def build_filename(self, str_rep: str, suffix=None) -> str:
        """
        Generate a filename in the cache.

        Parameters
        ----------
        str_rep: str
            Unique string representation of the item. Will be used to compute a hash.
        suffix: str. Default: None
            Optional file suffix, in order to allow filetype recognition by the name.

        Returns
        -------
            str
        """
        hashfile = os.path.join(
            self.folder, str(hashlib.sha256(str_rep.encode("ascii")).hexdigest())
        )
        if suffix is None:
            return hashfile
        else:
            if suffix.startswith("."):
                return hashfile + suffix
            else:
                return hashfile + "." + suffix


CACHE = Cache.instance()


class WarmupLevel(int, Enum):
    TEST = -1000
    INSTANCE = 1
    DATA = 5


class WarmupParam(NamedTuple):
    level: Union[int, WarmupLevel]
    fn: Callable
    is_factory: bool = False


class Warmup:

    _warmup_fns: List[WarmupParam] = []

    @staticmethod
    def fn_eql(wrapped_fn, original_fn):
        return wrapped_fn is original_fn or wrapped_fn.__wrapped__ is original_fn

    @classmethod
    def is_registered(cls, fn):
        return len([warmup_fn.fn
                    for warmup_fn in cls._warmup_fns
                    if cls.fn_eql(warmup_fn.fn, fn)]) > 0

    @classmethod
    def register_warmup_fn(cls, warmup_level: WarmupLevel = WarmupLevel.INSTANCE, *, is_factory=False):
        def outer(fn):
            if cls.is_registered(fn):
                raise WarmupRegException

            @wraps(fn)
            def inner(*args, **kwargs):
                return fn(*args, **kwargs)

            cls._warmup_fns.append(WarmupParam(warmup_level, inner, is_factory))
            return inner
        return outer

    @classmethod
    def deregister_warmup_fn(cls, original_fn):
        cls._warmup_fns = [
            warmup_fn for warmup_fn in cls._warmup_fns
            if not cls.fn_eql(warmup_fn.fn, original_fn)
        ]

    @classmethod
    def warmup(cls, warmup_level: WarmupLevel = WarmupLevel.INSTANCE, *, max_workers=4):
        all_fns = [warmup for warmup in cls._warmup_fns if warmup.level <= warmup_level]

        def call_fn(fn: WarmupParam):
            return_val = fn.fn()
            if not fn.is_factory:
                return
            for f in return_val:
                f()

        with Lock(CACHE.build_filename("lockfile", ".warmup")):
            with ThreadPoolExecutor(max_workers=max_workers) as ex:
                for _ in siibra_tqdm(
                    ex.map(
                        call_fn,
                        all_fns
                    ),
                    desc="Warming cache",
                    total=len(all_fns),
                ):
                    ...


try:
    from joblib import Memory
    jobmemory_path = Path(CACHE.folder) / "joblib"
    jobmemory_path.mkdir(parents=True, exist_ok=True)
    jobmemory = Memory(jobmemory_path, verbose=0)
    cache_user_fn = jobmemory.cache
except ImportError:
    from functools import lru_cache
    cache_user_fn = lru_cache



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, List, Union

import numpy as np
import pandas as pd

from . import tabular
from .. import anchor as _anchor
from ..feature import Compoundable
from ...core import region as _region
from ...commons import QUIET, siibra_tqdm
from ...locations import pointcloud
from ...retrieval.requests import HttpRequest
from ...retrieval.repositories import RepositoryConnector


class RegionalTimeseriesActivity(tabular.Tabular, Compoundable):
    """
    Datasets that provide regional activity over time.
    """

    _filter_attrs = ["modality", "cohort", "subject"]
    _compound_attrs = ["modality", "cohort"]

    def __init__(
        self,
        cohort: str,
        modality: str,
        regions: list,
        connector: RepositoryConnector,
        decode_func: Callable,
        filename: str,
        anchor: _anchor.AnatomicalAnchor,
        timestep: str,
        description: str = "",
        datasets: list = [],
        subject: str = "average",
        id: str = None,
        prerelease: bool = False,
    ):
        """
        """
        tabular.Tabular.__init__(
            self,
            modality=modality,
            description=description,
            anchor=anchor,
            datasets=datasets,
            data=None,  # lazy loading below
            id=id,
            prerelease=prerelease,
        )
        self.cohort = cohort.upper()
        if isinstance(connector, str) and connector:
            self._connector = HttpRequest(connector, decode_func)
        else:
            self._connector = connector
        self._filename = filename
        self._decode_func = decode_func
        self.regions = regions
        self._table = None
        self._subject = subject
        val, unit = timestep.split(" ")
        self.timestep = (float(val), unit)

    @property
    def subject(self):
        """Returns the subject identifiers for which the table represents."""
        return self._subject

    @property
    def name(self):
        return f"{self.subject} - " + super().name + f" cohort: {self.cohort}"

    @property
    def data(self) -> pd.DataFrame:
        """
        Returns a table as a pandas dataframe where the index is a timeseries.
        """
        if self._table is None:
            self._load_table()
        return self._table.copy()

    @classmethod
    def _merge_elements(
        cls,
        elements: List["RegionalTimeseriesActivity"],
        description: str,
        modality: str,
        anchor: _anchor.AnatomicalAnchor,
    ):
        assert len({f.cohort for f in elements}) == 1
        assert len({f.timestep for f in elements}) == 1
        merged = cls(
            cohort=elements[0].cohort,
            regions=elements[0].regions,
            connector=elements[0]._connector,
            decode_func=elements[0]._decode_func,
            filename="",
            timestep=" ".join(str(val) for val in elements[0].timestep),
            subject="average",
            description=description,
            modality=modality,
            anchor=anchor,
            **{"paradigm": "average"} if getattr(elements[0], "paradigm") else {}
        )
        if isinstance(elements[0]._connector, HttpRequest):
            getter = lambda elm: elm._connector.get()
        else:
            getter = lambda elm: elm._connector.get(elm._filename, decode_func=elm._decode_func)
        all_arrays = [
            getter(elm)
            for elm in siibra_tqdm(
                elements,
                total=len(elements),
                desc=f"Averaging {len(elements)} activity tables"
            )
        ]
        merged._table = elements[0]._arraylike_to_dataframe(
            np.stack(all_arrays).mean(0)
        )
        return merged

    def _load_table(self):
        """
        Extract the timeseries table.
        """
        if isinstance(self._connector, HttpRequest):
            array = self._connector.data
        else:
            array = self._connector.get(self._filename, decode_func=self._decode_func)
        self._table = self._arraylike_to_dataframe(array)

    def _arraylike_to_dataframe(self, array: Union[np.ndarray, pd.DataFrame]) -> pd.DataFrame:
        if not isinstance(array, np.ndarray):
            array = array.to_numpy()
        ncols = array.shape[1]
        table = pd.DataFrame(
            array,
            index=pd.TimedeltaIndex(
                np.arange(0, array.shape[0]) * self.timestep[0],
                unit=self.timestep[1],
                name="time"
            )
        )
        parcellations = self.anchor.represented_parcellations()
        assert len(parcellations) == 1
        parc = next(iter(parcellations))
        with QUIET:
            columnmap = {
                i: parc.get_region(regionname, allow_tuple=True)
                for i, regionname in enumerate(self.regions)
            }
        if len(columnmap) == ncols:
            remapper = {
                label - min(columnmap.keys()): region
                for label, region in columnmap.items()
            }
            table = table.rename(columns=remapper)

        return table

    def __str__(self):
        return self.name

    def compute_centroids(self, space):
        """
        Computes the list of centroid coordinates corresponding to
        dataframe columns, in the given reference space.

        Parameters
        ----------
        space: Space, str

        Returns
        -------
        list[tuple(float, float, float)]
        """
        result = []
        parcellations = self.anchor.represented_parcellations()
        assert len(parcellations) == 1
        parcmap = next(iter(parcellations)).get_map(space)
        all_centroids = parcmap.compute_centroids()
        for regionname in self.regions:
            region = parcmap.parcellation.get_region(regionname, allow_tuple=True)
            if isinstance(region, tuple):  # deal with sets of matched regions
                found = [c for r in region for c in r if c.name in all_centroids]
            else:
                found = [r for r in region if r.name in all_centroids]
            assert len(found) > 0
            result.append(
                tuple(pointcloud.PointCloud(
                    [all_centroids[r.name] for r in found], space=space
                ).centroid)
            )
        return result

    def plot(
        self, regions: List[Union[str, "_region.Region"]] = None, *args,
        backend="matplotlib", **kwargs
    ):
        """
        Create a bar plot of averaged timeseries data per region.

        Parameters
        ----------
        regions: List[str or Region]
        subject: str, default: None
            If None, returns the subject averaged table.
        args and kwargs:
            takes arguments and keyword arguments for the desired plotting
            backend.
        """
        if isinstance(regions, (str, _region.Region)):
            regions = [regions]
        if regions is None:
            regions = self.regions
        indices = [self.regions.index(r) for r in regions]
        table = self.data.iloc[:, indices]
        table.columns = [str(r) for r in table.columns]
        return table.mean().plot(kind="bar", *args, backend=backend, **kwargs)

    def plot_carpet(
        self, regions: List[Union[str, "_region.Region"]] = None, *args,
        backend="plotly", **kwargs
    ):
        """
        Create a carpet plot of the timeseries data per region.

        Parameters
        ----------
        regions: List[str or Region]
        subject: str, default: None
            If None, returns the subject averaged table.
        args and kwargs:
            takes arguments and keyword arguments for `plotly.express.imshow`
        """
        if backend != "plotly":
            raise NotImplementedError("Currently, carpet plot is only implemented with `plotly`.")
        if isinstance(regions, (str, _region.Region)):
            regions = [regions]
        if regions is None:
            regions = self.regions
        indices = [self.regions.index(r) for r in regions]
        table = self.data.iloc[:, indices].reset_index(drop=True)
        table.columns = [str(r) for r in table.columns]
        kwargs["title"] = kwargs.get("title", f"{self.modality}" + f" for subject={self.subject}")
        kwargs["labels"] = kwargs.get("labels", {
            "xlabel": self.data.index.to_numpy(dtype='timedelta64[ms]')}
        )
        from plotly.express import imshow
        return imshow(
            *args,
            table.T,
            **kwargs
        )


class RegionalBOLD(
    RegionalTimeseriesActivity,
    configuration_folder="features/tabular/activity_timeseries/bold",
    category="functional"
):
    """
    Blood-oxygen-level-dependent (BOLD) signals per region.
    """

    _filter_attrs = RegionalTimeseriesActivity._filter_attrs + ["paradigm"]
    _compound_attrs = RegionalTimeseriesActivity._compound_attrs + ["paradigm"]

    def __init__(self, paradigm: str, **kwargs):
        RegionalTimeseriesActivity.__init__(self, **kwargs)
        self.paradigm = paradigm

        # paradign is used to distinguish functional connectivity features from each other.
        assert self.paradigm, "RegionalBOLD must have paradigm defined!"

    @property
    def name(self):
        return super().name + f", paradigm: {self.paradigm}"



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Abstract base class for any kind of brain structure.
A brain structure is more general than a brain region. It refers to any object
defining a spatial extent in one or more reference spaces, and can thus be
used to compute intersections with other structures in space. For example,
a brain region is a structure which is at the same time an AtlasConcept. A
bounding box in MNI space is a structure, but not an AtlasConcept.
"""

from abc import ABC, abstractmethod
from typing import Tuple, Dict

from . import assignment, region as _region


class BrainStructure(ABC):
    """Abstract base class for types who can act as a location filter."""

    # cache assignment results at class level
    _ASSIGNMENT_CACHE: Dict[
        Tuple["BrainStructure", "BrainStructure"],
        "assignment.AnatomicalAssignment"
    ] = {}

    def intersects(self, other: "BrainStructure") -> bool:
        """
        Whether or not two BrainStructures have any intersection.
        """
        return self.intersection(other) is not None

    def __contains__(self, other: "BrainStructure") -> bool:
        return self.intersection(other) == other

    def __hash__(self) -> int:
        return hash(self.__repr__())

    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def intersection(self, other: "BrainStructure") -> "BrainStructure":
        """
        Return the intersection of two BrainStructures,
        ie. the other BrainStructure filtered by this BrainStructure.
        """
        pass

    def assign(self, other: "BrainStructure") -> assignment.AnatomicalAssignment:
        """
        Compute assignment of a BrainStructure to this filter.

        Parameters
        ----------
        other : Location or Region

        Returns
        -------
        assignment.AnatomicalAssignment or None
            None if there is no AssignmentQualification found.
        """
        # Two cases:
        # 1) self is location, other is location -> look at spatial intersection/relationship, do it here
        # 2) self is location, other is region -> get region map, then call again. do it here
        # If self is region -> Region overwrite this method, addressed there

        assert not isinstance(self, _region.Region)  # method is overwritten by Region!
        if (self, other) in self._ASSIGNMENT_CACHE:
            return self._ASSIGNMENT_CACHE[self, other]
        if (other, self) in self._ASSIGNMENT_CACHE:
            return self._ASSIGNMENT_CACHE[other, self].invert()

        if isinstance(other, _region.Region):
            inverse_assignment = other.assign(self)
            if inverse_assignment is None:
                self._ASSIGNMENT_CACHE[self, other] = None
            else:
                self._ASSIGNMENT_CACHE[self, other] = inverse_assignment.invert()
            return self._ASSIGNMENT_CACHE[self, other]
        else:  # other is a location object, just check spatial relationships
            qualification = None
            if self == other:
                qualification = assignment.Qualification.EXACT
            else:
                intersection = self.intersection(other)
                if intersection is not None:
                    if intersection == other:
                        qualification = assignment.Qualification.CONTAINS
                    elif intersection == self:
                        qualification = assignment.Qualification.CONTAINED
                    else:
                        qualification = assignment.Qualification.OVERLAPS
            if qualification is None:
                self._ASSIGNMENT_CACHE[self, other] = None
            else:
                self._ASSIGNMENT_CACHE[self, other] = assignment.AnatomicalAssignment(self, other, qualification)
        return self._ASSIGNMENT_CACHE[self, other]



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Matches BigBrain intensity profiles extracted by Wagstyl et al. to volumes."""

from typing import List
from os import path

import numpy as np
from scipy.spatial import KDTree

from . import query
from ..features import anchor as _anchor
from ..features.tabular import bigbrain_intensity_profile, layerwise_bigbrain_intensities
from ..features.image import CellbodyStainedSection, BigBrain1MicronPatch
from ..commons import logger, siibra_tqdm
from ..locations import point, pointcloud, location, experimental
from ..core import structure
from ..core.concept import get_registry
from ..retrieval import requests, cache
from ..retrieval.datasets import GenericDataset
from ..volumes import Volume, from_array


class WagstylProfileLoader:

    REPO = "https://github.com/kwagstyl/cortical_layers_tutorial/raw/main"
    PROFILES_FILE_LEFT = "data/profiles_left.npy"
    THICKNESSES_FILE_LEFT = "data/thicknesses_left.npy"
    MESH_FILE_LEFT = "data/gray_left_327680.surf.gii"
    _profiles = None
    _vertices = None
    _boundary_depths = None
    DATASET = GenericDataset(
        name="HIBALL workshop on cortical layers",
        contributors=[
            "Konrad Wagstyl",
            "Stéphanie Larocque",
            "Guillem Cucurull",
            "Claude Lepage",
            "Joseph Paul Cohen",
            "Sebastian Bludau",
            "Nicola Palomero-Gallagher",
            "Lindsay B. Lewis",
            "Thomas Funck",
            "Hannah Spitzer",
            "Timo Dickscheid",
            "Paul C. Fletcher",
            "Adriana Romero",
            "Karl Zilles",
            "Katrin Amunts",
            "Yoshua Bengio",
            "Alan C. Evans",
        ],
        url="https://github.com/kwagstyl/cortical_layers_tutorial/",
        description="Cortical profiles of BigBrain staining intensities computed by Konrad Wagstyl, "
        "as described in the publication 'Wagstyl, K., et al (2020). BigBrain 3D atlas of "
        "cortical layers: Cortical and laminar thickness gradients diverge in sensory and "
        "motor cortices. PLoS Biology, 18(4), e3000678. "
        "http://dx.doi.org/10.1371/journal.pbio.3000678."
        "The data is taken from the tutorial at "
        "https://github.com/kwagstyl/cortical_layers_tutorial. Each vertex is "
        "assigned to the regional map when queried.",
    )

    def __init__(self):
        if self._profiles is None:
            self.__class__._load()

    @property
    def profile_labels(self):
        return np.arange(0.0, 1.0, 1.0 / self._profiles.shape[1])

    @classmethod
    def _load(cls):
        # read thicknesses, in mm, and normalize by their last column which is the total thickness
        thickness = requests.HttpRequest(
            f"{cls.REPO}/{cls.THICKNESSES_FILE_LEFT}"
        ).data.T
        total_thickness = thickness[:, :-1].sum(
            1
        )  # last column is the computed total thickness
        valid = np.where(total_thickness > 0)[0]
        cls._boundary_depths = np.c_[
            np.zeros_like(valid),
            (thickness[valid, :] / total_thickness[valid, None]).cumsum(1),
        ]
        cls._boundary_depths[:, -1] = 1  # account for float calculation errors

        # find profiles with valid thickness
        profile_l_url = f"{cls.REPO}/{cls.PROFILES_FILE_LEFT}"
        if not path.exists(cache.CACHE.build_filename(profile_l_url)):
            logger.info(
                "First request to BigBrain profiles. Preprocessing the data "
                "now. This may take a little."
            )
        profiles_l_all = requests.HttpRequest(profile_l_url).data
        cls._profiles = profiles_l_all[valid, :]

        # read mesh vertices
        mesh_left = requests.HttpRequest(f"{cls.REPO}/{cls.MESH_FILE_LEFT}").data
        mesh_vertices = mesh_left.darrays[0].data
        cls._vertices = mesh_vertices[valid, :]

        logger.debug(f"{cls._profiles.shape[0]} BigBrain intensity profiles.")
        assert cls._vertices.shape[0] == cls._profiles.shape[0]

    def __len__(self):
        return self._vertices.shape[0]


cache.Warmup.register_warmup_fn()(lambda: WagstylProfileLoader._load())


class BigBrainProfileQuery(
    query.LiveQuery,
    args=[],
    FeatureType=bigbrain_intensity_profile.BigBrainIntensityProfile,
):

    def __init__(self):
        query.LiveQuery.__init__(self)

    def query(
        self, concept: structure.BrainStructure, **kwargs
    ) -> List[bigbrain_intensity_profile.BigBrainIntensityProfile]:
        loader = WagstylProfileLoader()
        mesh_vertices = pointcloud.PointCloud(loader._vertices, space="bigbrain")
        matched = concept.intersection(
            mesh_vertices
        )  # returns a reduced PointCloud with og indices as labels
        if matched is None:
            return []
        if isinstance(matched, point.Point):
            matched = pointcloud.from_points([matched])
        assert isinstance(matched, pointcloud.PointCloud)
        if isinstance(concept, location.Location):
            mesh_as_list = mesh_vertices.as_list()
            matched.labels = [mesh_as_list.index(v.coordinate) for v in matched]
        indices = matched.labels
        assert indices is not None
        features = []
        for i in indices:
            anchor = _anchor.AnatomicalAnchor(
                location=point.Point(loader._vertices[i], space="bigbrain"),
                region=str(concept),
                species="Homo sapiens",
            )
            prof = bigbrain_intensity_profile.BigBrainIntensityProfile(
                anchor=anchor,
                depths=loader.profile_labels,
                values=loader._profiles[i],
                boundaries=loader._boundary_depths[i],
            )
            prof.anchor._assignments[concept] = _anchor.AnatomicalAssignment(
                query_structure=concept,
                assigned_structure=concept,
                qualification=_anchor.Qualification.CONTAINED,
                explanation=f"Surface vertex of BigBrain cortical profile was filtered using {concept}",
            )
            prof.datasets = [WagstylProfileLoader.DATASET]
            features.append(prof)

        return features


class LayerwiseBigBrainIntensityQuery(
    query.LiveQuery,
    args=[],
    FeatureType=layerwise_bigbrain_intensities.LayerwiseBigBrainIntensities,
):

    def __init__(self):
        query.LiveQuery.__init__(self)

    def query(
        self, concept: structure.BrainStructure, **kwargs
    ) -> List[layerwise_bigbrain_intensities.LayerwiseBigBrainIntensities]:

        loader = WagstylProfileLoader()
        mesh_vertices = pointcloud.PointCloud(loader._vertices, space="bigbrain")
        matched = concept.intersection(
            mesh_vertices
        )  # returns a reduced PointCloud with og indices as labels if the concept is a region
        if matched is None:
            return []
        if isinstance(matched, point.Point):
            matched = pointcloud.from_points([matched])
        assert isinstance(matched, pointcloud.PointCloud)
        if isinstance(concept, location.Location):
            mesh_as_list = mesh_vertices.as_list()
            matched.labels = [mesh_as_list.index(v.coordinate) for v in matched]
        indices = matched.labels
        matched_profiles = loader._profiles[indices, :]
        boundary_depths = loader._boundary_depths[indices, :]
        # compute array of layer labels for all coefficients in profiles_left
        N = matched_profiles.shape[1]
        prange = np.arange(N)
        layer_labels = 7 - np.array(
            [
                [
                    np.array(
                        [
                            [(prange < T) * 1]
                            for i, T in enumerate((b * N).astype("int"))
                        ]
                    )
                    .squeeze()
                    .sum(0)
                ]
                for b in boundary_depths
            ]
        ).reshape((-1, 200))

        anchor = _anchor.AnatomicalAnchor(
            location=pointcloud.PointCloud(
                loader._vertices[indices, :], space="bigbrain"
            ),
            region=str(concept),
            species="Homo sapiens",
        )
        result = layerwise_bigbrain_intensities.LayerwiseBigBrainIntensities(
            anchor=anchor,
            means=[
                matched_profiles[layer_labels == layer].mean() for layer in range(1, 7)
            ],
            stds=[
                matched_profiles[layer_labels == layer].std() for layer in range(1, 7)
            ],
        )
        result.anchor._assignments[concept] = _anchor.AnatomicalAssignment(
            query_structure=concept,
            assigned_structure=concept,
            qualification=_anchor.Qualification.CONTAINED,
            explanation=f"Surface vertices of BigBrain cortical profiles were filtered using {concept}",
        )
        result.datasets = [WagstylProfileLoader.DATASET]

        return [result]


class BigBrain1MicronPatchQuery(
    query.LiveQuery, args=[], FeatureType=BigBrain1MicronPatch
):
    """
    Sample approximately orthogonal cortical image patches
    from BigBrain 1 micron sections, guided by an image volume
    in a supported reference space providing. The image
    volume is used as a weighted mask to extract patches
    along the cortical midsurface with nonzero weights in the
    input image.
    An optional lower_threshold can be used to narrow down
    the search
    The weight is stored with the resulting features.
    """

    def __init__(self, lower_threshold=0.):
        self.layermap = get_registry("Map").get("cortical layers bigbrain")
        self.lower_threshold = lower_threshold
        query.LiveQuery.__init__(self)

    def query(
        self, concept: structure.BrainStructure, **kwargs
    ) -> List[BigBrain1MicronPatch]:

        # make sure input is an image volume
        # TODO function should be extended to deal with other concepts as well
        if not isinstance(concept, Volume):
            logger.warning(
                "Querying BigBrain1MicronPatch features requires to "
                "query with an image volume."
            )
            return []

        # threshold image volume, if requested
        if self.lower_threshold > 0.0:
            logger.info(
                f"Applying lower threshold of {self.lower_threshold} "
                "for BigBrain 1 micron patch query."
            )
            img = concept.fetch()
            arr = np.asanyarray(img.dataobj)
            arr[arr < self.lower_threshold] = 0
            query_vol = from_array(arr, img.affine, space=concept.space, name="filtered volume")
        else:
            query_vol = concept
        bb_bbox = query_vol.get_boundingbox().warp('bigbrain')

        # find 1 micron BigBrain sections intersecting the thresholded volume
        sections: List[CellbodyStainedSection] = [
            s
            for s in CellbodyStainedSection._get_instances()
            if isinstance(s, CellbodyStainedSection)
            and s.get_boundingbox(clip=False).intersects(query_vol)
        ]
        if not sections:
            return []

        # extract relevant patches
        features = []
        for hemisphere in ["left", "right"]:

            # get layer 4 mesh in the hemisphere
            l4 = self.layermap.parcellation.get_region("4 " + hemisphere)
            l4mesh = self.layermap.fetch(l4, format="mesh")
            layerverts = {
                n: self.layermap.fetch(region=n, format="mesh")["verts"]
                for n in self.layermap.regions if hemisphere in n
            }
            l4verts = pointcloud.PointCloud(layerverts[l4.name], "bigbrain")
            if not l4verts.boundingbox.intersects(bb_bbox):
                continue

            # for each relevant BigBrain 1 micron section, intersect layer IV mesh
            # to obtain midcortex-locations, and build their orthogonal patches.
            # store the concept's value with the patch.
            vertex_tree = KDTree(layerverts[l4.name])
            for s in siibra_tqdm(
                sections, unit="sections", desc=f"Sampling patches in {hemisphere} hemisphere"
            ):

                # compute layer IV contour in the image plane
                imgplane = experimental.Plane.from_image(s)
                try:
                    contour_segments = imgplane.intersect_mesh(l4mesh)
                except AssertionError:
                    logger.error(f"Could not intersect with layer 4 mesh: {s.name}")
                    continue
                if len(contour_segments) == 0:
                    continue

                # score the contour points with the query image volume
                all_points = pointcloud.from_points(sum(map(list, contour_segments), []))
                all_probs = query_vol.evaluate_points(all_points)
                points_prob_lookup = {
                    pt.coordinate: prob
                    for pt, prob in zip(all_points, all_probs)
                    if prob >= self.lower_threshold
                }
                if len(points_prob_lookup) == 0:
                    continue

                # For each contour point,
                # - find the closest BigBrain layer surface vertex,
                # - build the profile of corresponding vertices across all layers
                # - project the profile to the image section
                # - determine the oriented patch along the profile
                _, indices = vertex_tree.query(np.array(list(points_prob_lookup.keys())))
                for prob, nnb in zip(points_prob_lookup.values(), indices):

                    prof = pointcloud.Contour(
                        [
                            layerverts[_][nnb]
                            for _ in self.layermap.regions
                            if hemisphere in _
                        ],
                        space=self.layermap.space,
                    )
                    patch = imgplane.get_enclosing_patch(prof)
                    if patch is None:
                        continue

                    anchor = _anchor.AnatomicalAnchor(
                        location=patch, species="Homo sapiens"
                    )
                    anchor._assignments[concept] = _anchor.AnatomicalAssignment(
                        query_structure=query_vol,
                        assigned_structure=s.anchor.volume,
                        qualification=_anchor.Qualification.CONTAINED
                    )
                    features.append(
                        BigBrain1MicronPatch(
                            patch=patch,
                            profile=prof,
                            section=s,
                            vertex=nnb,
                            relevance=prob,
                            anchor=anchor
                        )
                    )

        # return the patches sorted by relevance (ie. probability)
        return sorted(features, key=lambda p: p.relevance, reverse=True)



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A box defined by two farthest corner coordinates on a specific space."""

from itertools import product
import hashlib
from typing import TYPE_CHECKING, Union, Dict

import numpy as np

from . import location, point, pointcloud
from ..commons import logger
from ..exceptions import SpaceWarpingFailedError

if TYPE_CHECKING:
    from ..core.structure import BrainStructure
    from nibabel import Nifti1Image
    from ..core.space import Space


class BoundingBox(location.Location):
    """
    A 3D axis-aligned bounding box spanned by two 3D corner points.
    The box does not necessarily store the given points,
    instead it computes the real minimum and maximum points
    from the two corner points.
    """

    def __init__(
        self,
        point1,
        point2,
        space: Union[str, 'Space'] = None,
        minsize: float = None,
        sigma_mm=None
    ):
        """
        Construct a new bounding box spanned by two 3D coordinates
        in the given reference space.

        Parameters
        ----------
        point1 : Point or 3-tuple
            Startpoint given in mm of the given space
        point2 : Point or 3-tuple
            Endpoint given in mm of the given space
        space : reference space (id, name, or Space)
            The reference space
        minsize : float
            Minimum size along each dimension. If not None, the maxpoint will
            be adjusted to match the minimum size, if needed.
        sigma_mm : float, or list of float
            Optional standard deviation of the spanning point locations.
        """
        # TODO: allow to pass sigma for the points, if tuples
        location.Location.__init__(self, space)
        xyz1 = point.Point.parse(point1)
        xyz2 = point.Point.parse(point2)
        if sigma_mm is None:
            s1, s2 = 0, 0
        elif isinstance(sigma_mm, float):
            s1, s2 = sigma_mm, sigma_mm
        elif isinstance(sigma_mm, list):
            assert len(sigma_mm) == 2
            s1, s2 = sigma_mm
        else:
            raise ValueError(f"Cannot interpret sigma_mm parameter value {sigma_mm} for bounding box")
        self.sigma_mm = [s1, s2]
        self.minpoint = point.Point([min(xyz1[i], xyz2[i]) for i in range(3)], space, sigma_mm=s1)
        self.maxpoint = point.Point([max(xyz1[i], xyz2[i]) for i in range(3)], space, sigma_mm=s2)
        if minsize is not None:
            for d in range(3):
                if self.shape[d] < minsize:
                    self.maxpoint[d] = self.minpoint[d] + minsize

        if self.volume == 0:
            logger.debug(f"Zero-volume bounding box from points {point1} and {point2} in {self.space} space.")

    @property
    def id(self) -> str:
        return hashlib.md5(str(self).encode("utf-8")).hexdigest()

    @property
    def volume(self) -> float:
        """The volume of the boundingbox in mm^3"""
        return np.prod(self.shape)

    @property
    def center(self) -> 'point.Point':
        return self.minpoint + (self.maxpoint - self.minpoint) / 2

    @property
    def shape(self) -> float:
        """The distances of the diagonal points in each axis. (Accounts for sigma)."""
        return tuple(
            (self.maxpoint + self.maxpoint.sigma)
            - (self.minpoint - self.minpoint.sigma)
        )

    @property
    def is_planar(self) -> bool:
        return any(d == 0 for d in self.shape)

    def __str__(self):
        if self.space is None:
            return (
                f"Bounding box from ({','.join(f'{v:.2f}' for v in self.minpoint)})mm "
                f"to ({','.join(f'{v:.2f}' for v in self.maxpoint)})mm"
            )
        else:
            return (
                f"Bounding box from ({','.join(f'{v:.2f}' for v in self.minpoint)})mm "
                f"to ({','.join(f'{v:.2f}' for v in self.maxpoint)})mm in {self.space.name} space"
            )

    def intersection(self, other: 'BrainStructure', dims=[0, 1, 2]):
        """Computes the intersection of this bounding box with another one.

        Parameters
        ----------
        other: BrainStructure
        dims: List[int], default: all three
            Dimensions where the intersection should be computed
            (applies only to bounding boxes). Along dimensions not listed,
            the union is applied instead.
        """
        # TODO: process the sigma values o the points
        if isinstance(other, BoundingBox):
            try:
                return self._intersect_bbox(other, dims)
            except SpaceWarpingFailedError:
                return other._intersect_bbox(self, dims)  # TODO: check this mechanism carefully
        if isinstance(other, point.Point):
            warped = other.warp(self.space)
            return other if self.minpoint <= warped <= self.maxpoint else None
        if isinstance(other, pointcloud.PointCloud):
            points_inside = [p for p in other if self.intersects(p)]
            if len(points_inside) == 0:
                return None
            result = pointcloud.from_points(points_inside)
            return result[0] if len(result) == 1 else result  # if PointCloud has single point return as a Point

        return other.intersection(self)

    def _intersect_bbox(self, other: 'BoundingBox', dims=[0, 1, 2]):
        warped = other.warp(self.space)

        # Determine the intersecting bounsding box by sorting
        # the coordinates of both bounding boxes for each dimension,
        # and fetching the second and third coordinate after sorting.
        # If those belong to a minimum and maximum point,
        # no matter of which bounding box,
        # we have a nonzero intersection in that dimension.
        minpoints = [b.minpoint for b in (self, warped)]
        maxpoints = [b.maxpoint for b in (self, warped)]
        allpoints = minpoints + maxpoints
        result_minpt = []
        result_maxpt = []

        for dim in range(3):

            if dim not in dims:
                # do not intersect in this dimension, so take the union instead
                result_minpt.append(min(p[dim] for p in allpoints))
                result_maxpt.append(max(p[dim] for p in allpoints))
                continue

            A, B = sorted(allpoints, key=lambda P: P[dim])[1:3]
            if (A in maxpoints) or (B in minpoints):
                # no intersection in this dimension
                return None
            else:
                result_minpt.append(A[dim])
                result_maxpt.append(B[dim])

        if result_minpt == result_maxpt:
            return result_minpt

        bbox = BoundingBox(
            point1=point.Point(result_minpt, self.space),
            point2=point.Point(result_maxpt, self.space),
            space=self.space,
        )

        if bbox.volume == 0 and sum(cmin == cmax for cmin, cmax in zip(result_minpt, result_maxpt)) == 2:
            return None
        return bbox

    def _intersect_mask(self, mask: 'Nifti1Image', threshold=0):
        """Intersect this bounding box with an image mask. Returns None if they do not intersect.

        TODO process the sigma values o the points

        NOTE: The affine matrix of the image must be set to warp voxels
        coordinates into the reference space of this Bounding Box.
        """
        # nonzero voxel coordinates
        X, Y, Z = np.where(mask.get_fdata() > threshold)
        h = np.ones(len(X))

        # array of homogeneous physical nonzero voxel coordinates
        coords = np.dot(mask.affine, np.vstack((X, Y, Z, h)))[:3, :].T
        minpoint = [min(self.minpoint[i], self.maxpoint[i]) for i in range(3)]
        maxpoint = [max(self.minpoint[i], self.maxpoint[i]) for i in range(3)]
        minpt_voxel = np.dot(np.linalg.inv(mask.affine), np.r_[minpoint, 1])[:3]
        voxel_size = np.diff(
            np.dot(
                mask.affine,
                np.vstack((np.r_[minpt_voxel, 1], np.r_[minpt_voxel + 1, 1])).T
            )[:3, :]
        ).squeeze()
        # use voxel size for the test to intersect voxels properly with verythin bounding boxes
        inside = np.logical_and.reduce([
            coords >= minpoint - voxel_size / 2,
            coords <= maxpoint + voxel_size / 2
        ]).min(1)
        XYZ = coords[inside, :3]
        if XYZ.shape[0] == 0:
            return None
        elif XYZ.shape[0] == 1:  # reflect voxel size in resulting point set
            return self._intersect_bbox(
                point
                .Point(XYZ.flatten(), space=self.space, sigma_mm=voxel_size.max())
                .boundingbox
            )
        else:
            return self._intersect_bbox(
                pointcloud
                .PointCloud(XYZ, space=self.space, sigma_mm=voxel_size.max())
                .boundingbox
            )

    def clip(self, xyzmax, xyzmin=(0, 0, 0)):
        """
        Returns a new bounding box obtained by clipping at the given maximum coordinate.
        """
        return self.intersection(
            BoundingBox(
                point.Point(xyzmin, self.space), point.Point(xyzmax, self.space), self.space
            ),
            sigma_mm=[self.minpoint.sigma, self.maxpoint.sigma]
        )

    @property
    def corners(self):
        """
        Returns all 8 corners of the box as a pointcloud.

        Note
        ----
        x0, y0, z0 = self.minpoint
        x1, y1, z1 = self.maxpoint
        all_corners = [
            (x0, y0, z0),
            (x1, y0, z0),
            (x0, y1, z0),
            (x1, y1, z0),
            (x0, y0, z1),
            (x1, y0, z1),
            (x0, y1, z1),
            (x1, y1, z1)
        ]

        TODO: deal with sigma. Currently, returns the mean of min and max point.
        """
        xs, ys, zs = zip(self.minpoint, self.maxpoint)
        return pointcloud.PointCloud(
            coordinates=[[x, y, z] for x, y, z in product(xs, ys, zs)],
            space=self.space,
            sigma_mm=np.mean([self.minpoint.sigma, self.maxpoint.sigma])
        )

    def warp(self, space: Union[str, Dict, "Space"]):
        """Returns a new bounding box obtained by warping the
        min- and maxpoint of this one into the new target space.

        TODO process the sigma values o the points
        """
        from ..core.space import Space

        spaceobj = space if isinstance(space, Space) else Space.get_instance(space)
        if spaceobj == self.space:
            return self
        else:
            try:
                warped_corners = self.corners.warp(spaceobj)
            except SpaceWarpingFailedError:
                raise SpaceWarpingFailedError(f"Warping {str(self)} to {spaceobj.name} not successful.")
            return warped_corners.boundingbox

    def transform(self, affine: np.ndarray, space=None):
        """Returns a new bounding box obtained by transforming the
        min- and maxpoint of this one with the given affine matrix.

        TODO process the sigma values o the points

        Parameters
        ----------
        affine : numpy 4x4 ndarray
            affine matrix
        space : reference space (str, Space, or None)
            Target reference space which is reached after
            applying the transform. Note that the consistency
            of this cannot be checked and is up to the user.
        """
        from ..core.space import Space
        spaceobj = Space.get_instance(space)
        result = self.corners.transform(affine, spaceobj).boundingbox
        result.sigma_mm = [self.minpoint.sigma, self.maxpoint.sigma]  # TODO: error propagation
        return result

    def shift(self, offset):
        return self.__class__(
            point1=self.minpoint + offset,
            point2=self.maxpoint + offset,
            space=self.space,
            sigma_mm=[self.minpoint.sigma, self.maxpoint.sigma]
        )

    def zoom(self, ratio: float):
        """
        Create a new bounding box by zooming this one around its center.
        """
        c = self.center
        self.maxpoint - c
        return self.__class__(
            point1=(self.minpoint - c) * ratio + c,
            point2=(self.maxpoint - c) * ratio + c,
            space=self.space,
            sigma_mm=[self.minpoint.sigma * ratio, self.maxpoint.sigma * ratio]
        )

    def estimate_affine(self, space):
        """
        Computes an affine transform which approximates
        the nonlinear warping of the eight corner points
        to the desired target space.
        The transform is estimated using a least squares
        solution to A*x = b, where A is the matrix of
        point coefficients in the space of this bounding box,
        and b are the target coefficients in the given space
        after calling the nonlinear warping.
        """

        x0, y0, z0 = self.minpoint
        x1, y1, z1 = self.maxpoint

        # set of 8 corner points in source space
        corners1 = pointcloud.PointCloud(
            [
                (x0, y0, z0),
                (x0, y0, z1),
                (x0, y1, z0),
                (x0, y1, z1),
                (x1, y0, z0),
                (x1, y0, z1),
                (x1, y1, z0),
                (x1, y1, z1)
            ], self.space
        )

        # coefficient matrix from original points
        A = np.hstack(
            [
                [np.kron(np.eye(3), np.r_[tuple(c), 1])]
                for c in corners1
            ]).squeeze()

        # righthand side from warped points
        corners2 = corners1.warp(space)
        b = np.hstack(corners2.as_list())

        # least squares solution
        x, res, rank, s = np.linalg.lstsq(A, b, rcond=None)
        affine = np.vstack([x.reshape((3, 4)), np.array([0, 0, 0, 1])])

        # test
        errors = []
        for c1, c2 in zip(list(corners1), list(corners2)):
            errors.append(
                np.linalg.norm(
                    np.dot(affine, np.r_[tuple(c1), 1])
                    - np.r_[tuple(c2), 1]
                )
            )
        logger.debug(
            f"Average projection error under linear approximation "
            f"was {np.mean(errors):.2f} pixel"
        )

        return affine

    def __iter__(self):
        """Iterate the min- and maxpoint of this bounding box."""
        return iter((self.minpoint, self.maxpoint))

    def __eq__(self, other: 'BoundingBox'):
        if not isinstance(other, BoundingBox):
            return False
        return self.minpoint == other.minpoint and self.maxpoint == other.maxpoint

    def __hash__(self):
        return super().__hash__()


def _determine_bounds(masked_array: np.ndarray, background: float = 0.0):
    """
    Bounding box of nonzero (background) values in a 3D array.

    https://stackoverflow.com/questions/31400769/bounding-box-of-numpy-array
    """
    x = np.any(masked_array != background, axis=(1, 2))
    y = np.any(masked_array != background, axis=(0, 2))
    z = np.any(masked_array != background, axis=(0, 1))
    nzx, nzy, nzz = [np.where(v) for v in (x, y, z)]
    if any(len(nz[0]) == 0 for nz in [nzx, nzy, nzz]):
        # empty array
        return None
    xmin, xmax = nzx[0][[0, -1]]
    ymin, ymax = nzy[0][[0, -1]]
    zmin, zmax = nzz[0][[0, -1]]
    return np.array([[xmin, xmax + 1], [ymin, ymax + 1], [zmin, zmax + 1], [1, 1]])


def from_array(
    array: np.ndarray,
    background: Union[int, float] = 0.0,
    space: "Space" = None
) -> BoundingBox:
    """
    Find the bounding box of non-background values for any 3D array.

    Parameters
    ----------
    array: np.ndarray
    background: int or float, default: 0.0
    space: Space, default: None
    """
    bounds = _determine_bounds(array, background)
    return BoundingBox(point1=bounds[:3, 0], point2=bounds[:3, 1], space=space)



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handles multimodal data features and related queries."""

from typing import Union, TYPE_CHECKING, List, Dict, Type, Tuple, BinaryIO, Any, Iterator
from hashlib import md5
from collections import defaultdict
from zipfile import ZipFile
from abc import ABC, abstractmethod
from re import sub
from textwrap import wrap

from . import anchor as _anchor
from ..commons import logger, InstanceTable, siibra_tqdm, __version__
from ..core import concept, space, region, parcellation, structure
from ..volumes import volume

if TYPE_CHECKING:
    from ..retrieval.datasets import EbrainsDataset
    TypeDataset = EbrainsDataset


class ParseLiveQueryIdException(Exception):
    pass


class EncodeLiveQueryIdException(Exception):
    pass


class NotFoundException(Exception):
    pass


class ParseCompoundFeatureIdException(Exception):
    pass


_README_TMPL = """
Downloaded from siibra toolsuite.
siibra-python version: {version}

All related resources (e.g. doi, web resources) are categorized under publications.

Name
----
{name}

Description
-----------
{description}

Modality
--------
{modality}

{publications}
"""
_README_PUBLICATIONS = """
Publications
------------
{doi}

{ebrains_page}

{authors}

{publication_desc}

"""


class Feature:
    """
    Base class for anatomically anchored data features.
    """

    _SUBCLASSES: Dict[Type['Feature'], List[Type['Feature']]] = defaultdict(list)
    _CATEGORIZED: Dict[str, Type['InstanceTable']] = defaultdict(InstanceTable)

    category: str = None

    def __init__(
        self,
        modality: str,
        description: str,
        anchor: _anchor.AnatomicalAnchor,
        datasets: List['TypeDataset'] = [],
        id: str = None,
        prerelease: bool = False,
    ):
        """
        Parameters
        ----------
        modality: str
            A textual description of the type of measured information
        description: str
            A textual description of the feature.
        anchor: AnatomicalAnchor
        datasets : list
            list of datasets corresponding to this feature
        """
        self._modality_cached = modality
        self._description = description
        self._anchor_cached = anchor
        self.datasets = datasets
        self._id = id
        self._prerelease = prerelease

    @property
    def modality(self):
        # allows subclasses to implement lazy loading of the modality
        return self._modality_cached

    @property
    def anchor(self):
        # allows subclasses to implement lazy loading of an anchor
        return self._anchor_cached

    def __init_subclass__(cls, configuration_folder=None, category=None, **kwargs):

        # Feature.SUBCLASSES serves as an index where feature class inheritance is cached. When users
        # queries a branch on the hierarchy, all children will also be queried. There are usecases where
        # such behavior is not desired (e.g. ProxyFeature, which wraps livequery features id to capture the
        # query context).
        if "ProxyFeature" in cls.__name__:
            return

        # extend the subclass lists
        # Iterate over all mro, not just immediate base classes
        for BaseCls in cls.__mro__:
            # some base classes may not be sub class of feature, ignore these
            if not issubclass(BaseCls, Feature):
                continue
            cls._SUBCLASSES[BaseCls].append(cls)

        cls._live_queries = []
        cls._preconfigured_instances = None
        cls._configuration_folder = configuration_folder
        cls.category = category
        if category is not None:
            cls._CATEGORIZED[category].add(cls.__name__, cls)
        return super().__init_subclass__(**kwargs)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}(id='{self.id}', name='{self.name}')>"

    @classmethod
    def _get_subclasses(cls):
        return {Cls.__name__: Cls for Cls in cls._SUBCLASSES}

    @property
    def description(self):
        """Allows subclasses to overwrite the description with a function call."""
        if self._description:
            return self._description
        for ds in self.datasets:
            if ds.description:
                return ds.description
        return ''

    @property
    def LICENSE(self) -> str:
        licenses = []
        for ds in self.datasets:
            if ds.LICENSE is None or ds.LICENSE == "No license information is found.":
                continue
            if isinstance(ds.LICENSE, str):
                licenses.append(ds.LICENSE)
            if isinstance(ds.LICENSE, list):
                licenses.extend(ds.LICENSE)
        if len(licenses) == 0:
            logger.warning("No license information is found.")
            return ""
        if len(licenses) > 1:
            logger.info("Found multiple licenses corresponding to datasets.")
        return '\n'.join(licenses)

    @property
    def urls(self) -> List[str]:
        """The list of URLs (including DOIs) associated with this feature."""
        return [
            url.get("url")
            for ds in self.datasets
            for url in ds.urls
        ]

    @property
    def authors(self):
        return [
            contributor['name']
            for ds in self.datasets
            for contributor in ds.contributors
        ]

    @property
    def name(self):
        """Returns a short human-readable name of this feature."""
        readable_class_name = sub("([a-z])([A-Z])", r"\g<1> \g<2>", self.__class__.__name__)
        name_ = sub("([b,B]ig [b,B]rain)", "BigBrain", readable_class_name)
        return name_ if not self._prerelease else f"[PRERELEASE] {name_}"

    @classmethod
    def _get_instances(cls, **kwargs) -> List['Feature']:
        """
        Retrieve objects of a particular feature subclass.
        Objects can be preconfigured in the configuration,
        or delivered by Live queries.
        """
        if not hasattr(cls, "_preconfigured_instances"):
            return []

        if cls._preconfigured_instances is not None:
            return cls._preconfigured_instances

        if cls._configuration_folder is None:
            cls._preconfigured_instances = []
            return cls._preconfigured_instances

        from ..configuration.configuration import Configuration
        conf = Configuration()
        Configuration.register_cleanup(cls._clean_instances)
        if cls._configuration_folder not in conf.folders:
            logger.debug(f"{cls._configuration_folder} is not in current configuration")
            return []

        cls._preconfigured_instances = [
            o for o in conf.build_objects(cls._configuration_folder)
            if isinstance(o, cls)
        ]
        logger.debug(
            f"Built {len(cls._preconfigured_instances)} preconfigured {cls.__name__} "
            f"objects from {cls._configuration_folder}."
        )
        return cls._preconfigured_instances

    def plot(self, *args, **kwargs):
        """Feature subclasses override this with their customized plot methods."""
        raise NotImplementedError("Generic feature class does not have a standardized plot.")

    @classmethod
    def _clean_instances(cls):
        """ Removes all instantiated object instances"""
        cls._preconfigured_instances = None

    def matches(
        self,
        concept: Union[structure.BrainStructure, space.Space],
    ) -> bool:
        """
        Match the features anatomical anchor against the given query concept.
        Record the most recently matched concept for inspection by the caller.
        """
        # TODO: storing the last matched concept. It is not ideal, might cause problems in multithreading
        if self.anchor and self.anchor.matches(concept):
            self.anchor._last_matched_concept = concept
            return True
        self.anchor._last_matched_concept = None
        return False

    @property
    def last_match_result(self):
        "The result of the last anchor comparison to a BrainStructure."
        return None if self.anchor is None else self.anchor.last_match_result

    @property
    def last_match_description(self):
        "The description of the last anchor comparison to a BrainStructure."
        return "" if self.anchor is None else self.anchor.last_match_description

    @property
    def id(self):
        if self._id:
            return self._id

        if self._prerelease:
            name_ = self.name.replace("[PRERELEASE] ", "")
        else:
            name_ = self.name

        prefix = ''
        for ds in self.datasets:
            if hasattr(ds, "id"):
                prefix = ds.id + '--'
                break
        return prefix + md5(
            f"{name_} - {self.anchor}".encode("utf-8")
        ).hexdigest()

    def _to_zip(self, fh: ZipFile):
        """
        Internal implementation. Subclasses can override but call super()._to_zip(fh).
        This allows all classes in the __mro__ to have the opportunity to append files
        of interest.
        """
        if isinstance(self, Compoundable) and "README.md" in fh.namelist():
            return
        ebrains_page = "\n".join(
            {ds.ebrains_page for ds in self.datasets if getattr(ds, "ebrains_page", None)}
        )
        doi = "\n".join({
            u.get("url")
            for ds in self.datasets if ds.urls
            for u in ds.urls
        })
        authors = ", ".join({
            cont.get('name')
            for ds in self.datasets if ds.contributors
            for cont in ds.contributors
        })
        publication_desc = "\n".join({ds.description for ds in self.datasets})
        if (ebrains_page or doi) and authors:
            publications = _README_PUBLICATIONS.format(
                ebrains_page="EBRAINS page\n" + ebrains_page if ebrains_page else "",
                doi="DOI\n" + doi if doi else "",
                authors="Authors\n" + authors if authors else "",
                publication_desc="Publication description\n" + publication_desc if publication_desc else ""
            )
        else:
            publications = "Note: could not obtain any publication information. The data may not have been published yet."
        fh.writestr(
            "README.md",
            _README_TMPL.format(
                version=__version__,
                name=self.name,
                description=self.description,
                modality=self.modality,
                publications=publications
            )
        )

    def to_zip(self, filelike: Union[str, BinaryIO]):
        """
        Export as a zip archive.

        Parameters
        ----------
        filelike: str or path
            Filelike to write the zip file. User is responsible to ensure the
            correct extension (.zip) is set.
        """
        fh = ZipFile(filelike, "w")
        self._to_zip(fh)
        fh.close()

    @staticmethod
    def _serialize_query_context(feat: 'Feature', concept: concept.AtlasConcept) -> str:
        """
        Serialize feature from livequery and query context.

        It is currently impossible to retrieve a livequery with a generic UUID.
        As such, the query context (e.g. region, space or parcellation) needs to
        be encoded in the id.

        Whilst it is possible to (de)serialize *any* queries, the method is setup to only serialize
        livequery features.

        The serialized livequery id follows the following pattern:

        <livequeryid_version>::<feature_cls_name>::<query_context>::<unserialized_id>

        Where:

        - livequeryid_version: version of the serialization. (e.g. lq0)
        - feature_cls_name: class name to query. (e.g. BigBrainIntensityProfile)
        - query_context: string to retrieve atlas concept in the query context. Can be one of the following:
            - s:<space_id>
            - p:<parcellation_id>
            - p:<parcellation_id>::r:<region_id>
        - unserialized_id: id prior to serialization

        See test/features/test_feature.py for tests and usages.
        """
        if not hasattr(feat.__class__, '_live_queries'):
            raise EncodeLiveQueryIdException(f"generate_livequery_featureid can only be used on live queries, but {feat.__class__.__name__} is not.")

        encoded_c = Feature._encode_concept(concept)

        return f"lq0::{feat.__class__.__name__}::{encoded_c}::{feat.id}"

    @classmethod
    def _deserialize_query_context(cls, feature_id: str) -> Tuple[Type['Feature'], concept.AtlasConcept, str]:
        """
        Deserialize id into query context.

        See docstring of serialize_query_context for context.
        """
        lq_version, *rest = feature_id.split("::")
        if lq_version != "lq0":
            raise ParseLiveQueryIdException("livequery id must start with lq0::")

        clsname, *concepts, fid = rest

        Features = cls._parse_featuretype(clsname)

        if len(Features) == 0:
            raise ParseLiveQueryIdException(f"classname {clsname!r} could not be parsed correctly. {feature_id!r}")
        F = Features[0]

        concept = cls._decode_concept(concepts)

        return (F, concept, fid)

    @staticmethod
    def _encode_concept(concept: concept.AtlasConcept):
        from ..locations import Location
        encoded_c = []
        if isinstance(concept, space.Space):
            encoded_c.append(f"s:{concept.id}")
        elif isinstance(concept, parcellation.Parcellation):
            encoded_c.append(f"p:{concept.id}")
        elif isinstance(concept, region.Region):
            encoded_c.append(f"p:{concept.parcellation.id}")
            encoded_c.append(f"r:{concept.name}")
        elif isinstance(concept, volume.Volume):
            encoded_c.append(f"v:{concept.name}")
        elif isinstance(concept, Location):
            encoded_c.append(f"loc:{Location}")

        if len(encoded_c) == 0:
            raise EncodeLiveQueryIdException("no concept is encoded")

        return '::'.join(encoded_c)

    @classmethod
    def _decode_concept(cls, concepts: List[str]) -> concept.AtlasConcept:
        # choose exception to divert try-except correctly
        if issubclass(cls, CompoundFeature):
            exception = ParseCompoundFeatureIdException
        else:
            exception = ParseLiveQueryIdException

        concept = None
        for c in concepts:
            if c.startswith("s:"):
                if concept is not None:
                    raise exception("Conflicting spec.")
                concept = space.Space.registry()[c.replace("s:", "")]
            if c.startswith("p:"):
                if concept is not None:
                    raise exception("Conflicting spec.")
                concept = parcellation.Parcellation.registry()[c.replace("p:", "")]
            if c.startswith("r:"):
                if concept is None:
                    raise exception("region has been encoded, but parcellation has not been populated in the encoding, {feature_id!r}")
                if not isinstance(concept, parcellation.Parcellation):
                    raise exception("region has been encoded, but previous encoded concept is not parcellation")
                concept = concept.get_region(c.replace("r:", ""))

        if concept is None:
            raise ParseLiveQueryIdException("concept was not populated in feature id")
        return concept

    @classmethod
    def _parse_featuretype(cls, feature_type: str) -> List[Type['Feature']]:
        ftypes = sorted({
            feattype
            for FeatDataCls, feattypes in cls._SUBCLASSES.items()
            if all(w.lower() in FeatDataCls.__name__.lower() for w in feature_type.split())
            for feattype in feattypes
            if getattr(feattype, 'category')
        }, key=lambda t: t.__name__)
        return sorted(ftypes, key=lambda t: t.__name__)

    @classmethod
    def _livequery(cls, concept: Union[region.Region, parcellation.Parcellation, space.Space], **kwargs) -> List['Feature']:
        if not hasattr(cls, "_live_queries"):
            return []

        live_instances = []
        for QueryType in cls._live_queries:
            argstr = f" ({', '.join('='.join(map(str, _)) for _ in kwargs.items())})" \
                if len(kwargs) > 0 else ""
            logger.debug(
                f"Running live query for {QueryType.feature_type.__name__} "
                f"objects linked to {str(concept)}{argstr}"
            )
            q = QueryType(**kwargs)
            if isinstance(concept, space.Space):
                features = q.query(concept.get_template())
            else:
                features = q.query(concept)
            live_instances.extend(
                Feature._wrap_livequery_feature(f, Feature._serialize_query_context(f, concept))
                for f in features
            )

        return live_instances

    @classmethod
    def _match(
        cls,
        concept: Union[structure.BrainStructure, space.Space],
        feature_type: Union[str, Type['Feature'], List['Feature']],
        **kwargs
    ) -> List['Feature']:
        """
        Retrieve data features of the requested feature type (i.e. modality).
        This will
        - call Feature.match(concept) for any registered preconfigured features
        - run any registered live queries
        The preconfigured and live query instances are merged and returned as a list.

        If multiple feature types are given, recurse for each of them.


        Parameters
        ----------
        concept: AtlasConcept
            An anatomical concept, typically a brain region or parcellation.
        feature_type: subclass of Feature, str
            specifies the type of features ("modality")
        """
        if isinstance(feature_type, list):
            # a list of feature types is given, collect match results on those
            ftypes = list(dict.fromkeys(
                cls._parse_featuretype(ft) if isinstance(ft, str) else ft
                for ft in feature_type
            ))
            return list(dict.fromkeys(
                sum((
                    cls._match(concept, t, **kwargs) for t in ftypes
                ), [])
            ))

        if isinstance(feature_type, str):
            # feature type given as a string. Decode the corresponding class.
            # Some string inputs, such as connectivity, may hit multiple matches.
            ftype_candidates = cls._parse_featuretype(feature_type)
            if len(ftype_candidates) == 0:
                raise ValueError(
                    f"feature_type {str(feature_type)} did not match with any "
                    f"features. Available features are: {', '.join(cls._SUBCLASSES.keys())}"
                )
            logger.info(
                f"'{feature_type}' decoded as feature type/s: "
                f"{[c.__name__ for c in ftype_candidates]}."
            )
            return cls._match(concept, ftype_candidates, **kwargs)

        assert issubclass(feature_type, Feature)

        # At this stage, no recursion is needed.
        # We expect a specific supported feature type is to be matched now.
        if not isinstance(concept, (structure.BrainStructure, space.Space)):
            raise ValueError(
                f"{concept.__class__.__name__} cannot be used for feature queries as it is not a `BrainStructure` or a `Space` type."
            )

        # Collect any preconfigured instances of the requested feature type
        # which match the query concept
        if feature_type.category == "generic":
            modality = kwargs.pop("modality", "")
            feature_type_instances = [
                f
                for f in feature_type._get_instances()
                if all(
                    w in f.modality.lower()
                    for w in modality.lower().split(" ")
                )
            ]
        else:
            feature_type_instances = [
                instance
                for f_type in cls._SUBCLASSES[feature_type]
                for instance in f_type._get_instances()
            ]

        matching_preconf_instances = [
            f for f in siibra_tqdm(
                feature_type_instances,
                desc=f"Matching {feature_type.__name__} to {concept}",
                total=len(feature_type_instances),
                disable=(not feature_type_instances)
            )
            if f.matches(concept)
        ]

        # Then run any registered live queries for the requested feature type
        # with the query concept.
        live_instances = feature_type._livequery(concept, **kwargs)

        results = sorted(
            dict.fromkeys(matching_preconf_instances + live_instances),  # to remove duplicates
            key=lambda f: min(f.last_match_result) if f.last_match_result else False,  # to order according to assignmnent ranking
        )
        return CompoundFeature._compound(results, concept)

    @classmethod
    def _get_instance_by_id(cls, feature_id: str, **kwargs):
        try:
            return CompoundFeature._get_instance_by_id(feature_id, **kwargs)
        except ParseCompoundFeatureIdException:
            pass

        try:
            F, concept, fid = cls._deserialize_query_context(feature_id)
            return [
                f
                for f in F._livequery(concept, **kwargs)
                if f.id == fid or f.id == feature_id
            ][0]
        except ParseLiveQueryIdException:
            candidates = [
                inst
                for Cls in Feature._SUBCLASSES[Feature]
                for inst in Cls._get_instances()
                if inst.id == feature_id
            ]
            if len(candidates) == 0:
                raise NotFoundException(f"No feature instance with {feature_id} found.")
            if len(candidates) == 1:
                return candidates[0]
            else:
                raise RuntimeError(
                    f"Multiple feature instance match {feature_id}",
                    [c.name for c in candidates]
                )
        except IndexError:
            raise NotFoundException

    @staticmethod
    def _wrap_livequery_feature(feature: 'Feature', fid: str):
        """
        Wrap live query features, override only the id attribute.

        Some features do not have setters for the id property. The ProxyFeature class
        allow the id property to be overridden without touching the underlying class.

        See docstring of serialize_query_context for further context.
        """

        # if you change the name of this class, change the string in Feature.__init_subclass__
        class ProxyFeature(feature.__class__):

            # override __class__ property
            # some instances of features accesses inst.__class__
            @property
            def __class__(self):
                return self.inst.__class__

            def __init__(self, inst: Feature, fid: str):
                self.inst = inst
                self.fid = fid
                self.category = inst.category

            def __str__(self) -> str:
                return self.inst.__str__()

            def __repr__(self) -> str:
                return self.inst.__repr__()

            @property
            def id(self):
                return self.fid

            def __getattr__(self, __name: str):
                return getattr(self.inst, __name)

        return ProxyFeature(feature, fid)


class Compoundable(ABC):
    """
    Base class for structures which allow compounding.
    Determines the necessary grouping and compounding attributes.
    """
    _filter_attrs = []  # the attributes to filter this instance of feature
    _compound_attrs = []  # `compound_key` has to be created from `filter_attributes`

    def __init_subclass__(cls, **kwargs):
        assert len(cls._filter_attrs) > 0, "All compoundable classes have to have `_filter_attrs` defined."
        assert len(cls._compound_attrs) > 0, "All compoundable classes have to have `_compound_attrs` defined."
        assert all(attr in cls._filter_attrs for attr in cls._compound_attrs), "`_compound_attrs` must be a subset of `_filter_attrs`."
        cls._indexing_attrs = [
            attr
            for attr in cls._filter_attrs
            if attr not in cls._compound_attrs
        ]
        return super().__init_subclass__(**kwargs)

    def __init__(self):
        assert all(hasattr(self, attr) for attr in self._filter_attrs), "`_filter_attrs` can only consist of the attributes of the class."

    @property
    def filter_attributes(self) -> Dict[str, Any]:
        """
        Attributes that help distinguish or combine features of the same type
        among others.
        """
        return {attr: getattr(self, attr) for attr in self._filter_attrs}

    @property
    def _compound_key(self) -> Tuple[Any]:
        """
        A tuple of values that define the basis for compounding elements of
        the same type.
        """
        return tuple([self.filter_attributes[attr] for attr in self._compound_attrs])

    @property
    def _element_index(self) -> Any:
        """
        Unique index of this compoundable feature as an element of the Compound.
        Must be hashable.
        """
        index_ = [self.filter_attributes[attr] for attr in self._indexing_attrs]
        index = index_[0] if len(index_) == 1 else tuple(index_)
        assert hash(index), "`_element_index` of a compoundable must be hashable."
        return index

    @classmethod
    def _merge_anchors(cls, anchors: List[_anchor.AnatomicalAnchor]):
        return sum(anchors)

    @classmethod
    @abstractmethod
    def _merge_elements(
        cls,
        elements,
        description: str,
        modality: str,
        anchor: _anchor.AnatomicalAnchor
    ) -> Feature:
        """
        Compute the merge data and create a merged instance from a set of
        elements of this class. This will be used by CompoundFeature to
        create the aggregated data and plot it. For example, to compute an
        average connectivity matrix from a set of subfeatures, we create a
        RegionalConnectivty feature.
        """
        raise NotImplementedError


class CompoundFeature(Feature):
    """
    A compound aggregating multiple features of the same type, forming its
    elements. The anatomical anchors and data of the features is merged.
    Features need to subclass "Compoundable" to allow aggregation
    into a compound feature.
    """

    def __init__(
        self,
        elements: List['Feature'],
        queryconcept: Union[region.Region, parcellation.Parcellation, space.Space],
    ):
        """
        A compound of several features of the same type with an anchor created
        as a sum of adjoinable anchors.
        """
        self._feature_type = elements[0].__class__
        assert all(isinstance(f, self._feature_type) for f in elements), NotImplementedError("Cannot compound features of different types.")
        self.category = elements[0].category  # same feature types have the same category
        assert issubclass(self._feature_type, Compoundable), NotImplementedError(f"Cannot compound {self._feature_type}.")

        modality = elements[0].modality
        assert all(f.modality == modality for f in elements), NotImplementedError("Cannot compound features of different modalities.")

        compound_keys = {element._compound_key for element in elements}
        assert len(compound_keys) == 1, ValueError(
            "Only features with identical compound_key can be aggregated."
        )
        self._compounding_attributes = {
            attr: elements[0].filter_attributes[attr]
            for attr in elements[0]._compound_attrs
        }

        self._elements = {f._element_index: f for f in elements}
        assert len(self._elements) == len(elements), RuntimeError(
            "Element indices should be unique to each element within a CompoundFeature."
        )

        Feature.__init__(
            self,
            modality=modality,
            description="\n".join({f.description for f in elements}),
            anchor=self._feature_type._merge_anchors([f.anchor for f in elements]),
            datasets=list(dict.fromkeys([ds for f in elements for ds in f.datasets])),
            prerelease=all(f._prerelease for f in elements),
        )
        self._queryconcept = queryconcept
        self._merged_feature_cached = None

    def __getattr__(self, attr: str) -> Any:
        """Expose compounding attributes explicitly."""
        if attr in self._compounding_attributes:
            return self._compounding_attributes[attr]
        if hasattr(self._feature_type, attr):
            raise AttributeError(
                f"{self.__class__.__name__} does not have access to '{attr}' "
                "since it does not have the same value for all its elements."
            )
        raise AttributeError(
            f"{self.__class__.__name__} or {self._feature_type.__name__} have no attribute {attr}."
        )

    def __dir__(self):
        return super().__dir__() + list(self._compounding_attributes.keys())

    def plot(self, *args, **kwargs):
        kwargs["title"] = "(Derived data: averaged)\n" + kwargs.get(
            "title",
            "\n".join(wrap(self.name, kwargs.pop("textwrap", 40)))
        )
        return self._get_merged_feature().plot(*args, **kwargs)

    def _get_merged_feature(self) -> Feature:
        if self._merged_feature_cached is None:
            logger.info(f"{self.__class__.__name__}.data averages the data of each element.")
            assert issubclass(self.feature_type, Compoundable)
            self._merged_feature_cached = self.feature_type._merge_elements(
                elements=self.elements,
                modality=self.modality,
                description=self.description,
                anchor=self.anchor
            )
        return self._merged_feature_cached

    @property
    def data(self):
        return self._get_merged_feature().data

    @property
    def indexing_attributes(self) -> Tuple[str]:
        "The attributes determining the index of this CompoundFeature's elements."
        return tuple(self.elements[0]._indexing_attrs)

    @property
    def elements(self):
        """Features that make up the compound feature."""
        return list(self._elements.values())

    @property
    def indices(self):
        """Unique indices to features making up the CompoundFeature."""
        return list(self._elements.keys())

    @property
    def feature_type(self) -> Type:
        """Feature type of the elements forming the CompoundFeature."""
        return self._feature_type

    @property
    def name(self) -> str:
        """Returns a short human-readable name of this feature."""
        readable_feature_type = sub(
            "([b,B]ig [b,B]rain)", "BigBrain",
            sub("([a-z])([A-Z])", r"\g<1> \g<2>", self.feature_type.__name__)
        )
        groupby = ', '.join([
            f"{k}: {v}"
            for k, v in self._compounding_attributes.items()
            if k != 'modality'
        ])
        cf_name = f"{len(self)} {readable_feature_type} features{f' {groupby}' if groupby else ''}"
        return cf_name if not self._prerelease else f"[PRERELEASE] {cf_name}"

    @property
    def id(self) -> str:
        if self._prerelease:
            name_ = self.name.replace("[PRERELEASE] ", "")
        else:
            name_ = self.name
        return "::".join((
            "cf0",
            f"{self._feature_type.__name__}",
            self._encode_concept(self._queryconcept),
            self.datasets[0].id if self.datasets else "nodsid",
            md5(name_.encode("utf-8")).hexdigest()
        ))

    def __iter__(self) -> Iterator['Feature']:
        """Iterate over subfeatures"""
        return self.elements.__iter__()

    def __len__(self):
        """Number of subfeatures making the CompoundFeature"""
        return len(self._elements)

    def __getitem__(self, index: Any):
        """Get the nth element in the compound."""
        return self.elements[index]

    def get_element(self, index: Any):
        """Get the element with its unique index in the compound."""
        try:
            return self._elements[index]
        except Exception:
            raise IndexError(f"No feature with index '{index}' in this compound.")

    @classmethod
    def _compound(
        cls,
        features: List['Feature'],
        queryconcept: Union[region.Region, parcellation.Parcellation, space.Space]
    ) -> List['CompoundFeature']:
        """
        Compound features of the same the same type based on their `_compound_key`.

        If there are features that are not of type `Compoundable`, they are
        returned as is.

        Parameters
        ----------
        features: List[Feature]
            Feature instances to be compounded.
        queryconcept:
            AtlasConcept used for the query.

        Returns
        -------
        List[CompoundFeature | Feature]
        """
        non_compound_features = []
        grouped_features = defaultdict(list)
        for f in features:
            if isinstance(f, Compoundable):
                grouped_features[f._compound_key].append(f)
                continue
            non_compound_features.append(f)
        return non_compound_features + [
            cls(fts, queryconcept)
            for fts in grouped_features.values() if fts
        ]

    @classmethod
    def _get_instance_by_id(cls, feature_id: str, **kwargs):
        """
        Use the feature id to obtain the same feature instance.

        Parameters
        ----------
        feature_id : str

        Returns
        -------
        CompoundFeature

        Raises
        ------
        ParseCompoundFeatureIdException
            If no or multiple matches are found. Or id is not fitting to
            compound features.
        """
        if not feature_id.startswith("cf0::"):
            raise ParseCompoundFeatureIdException("CompoundFeature id must start with cf0::")
        cf_version, clsname, *queryconcept, dsid, fid = feature_id.split("::")
        assert cf_version == "cf0"
        candidates = [
            f
            for f in Feature._match(
                concept=cls._decode_concept(queryconcept),
                feature_type=clsname
            )
            if f.id == fid or f.id == feature_id
        ]
        if candidates:
            if len(candidates) == 1:
                return candidates[0]
            else:
                raise ParseCompoundFeatureIdException(
                    f"The query with id '{feature_id}' have resulted multiple instances.")
        else:
            raise ParseCompoundFeatureIdException

    def _to_zip(self, fh: ZipFile):
        super()._to_zip(fh)
        for idx, element in siibra_tqdm(self._elements.items(), desc="Exporting elements", unit="element"):
            if '/' in str(idx):
                logger.warning(f"'/' will be replaced with ' ' of the file for element with index {idx}")
            filename = '/'.join([
                str(i).replace('/', ' ')
                for i in (idx if isinstance(idx, tuple) else [idx])
            ])
            fh.writestr(f"{self.feature_type.__name__}/{filename}.csv", element.data.to_csv())



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Constants, functions, and classes used commonly across siibra."""

import os
import re
from enum import Enum
import logging
from dataclasses import dataclass
from hashlib import md5
from uuid import UUID
import math
from typing import Generic, Iterable, Iterator, List, TypeVar, Union, Dict, Generator, Tuple
try:
    from typing import TypedDict
except ImportError:
    # support python 3.7
    from typing_extensions import TypedDict

from tqdm import tqdm
import numpy as np
import pandas as pd
from nibabel import Nifti1Image
from nilearn.image import resample_to_img
from skimage.filters import gaussian


logging.addLevelName(21, "INFO_WO_PROGRESS_BARS")
logger = logging.getLogger(__name__.split(os.path.extsep)[0])
ch = logging.StreamHandler()
formatter = logging.Formatter("[{name}:{levelname}] {message}", style="{")
ch.setFormatter(formatter)
logger.addHandler(ch)


ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
HBP_AUTH_TOKEN = os.getenv("HBP_AUTH_TOKEN")
KEYCLOAK_CLIENT_ID = os.getenv("KEYCLOAK_CLIENT_ID")
KEYCLOAK_CLIENT_SECRET = os.getenv("KEYCLOAK_CLIENT_SECRET")
SIIBRA_CACHEDIR = os.getenv("SIIBRA_CACHEDIR")
SIIBRA_LOG_LEVEL = os.getenv("SIIBRA_LOG_LEVEL", "INFO")
SIIBRA_USE_CONFIGURATION = os.getenv("SIIBRA_USE_CONFIGURATION")
SIIBRA_USE_LOCAL_SNAPSPOT = os.getenv("SIIBRA_USE_LOCAL_SNAPSPOT")
SKIP_CACHEINIT_MAINTENANCE = os.getenv("SKIP_CACHEINIT_MAINTENANCE")
SIIBRA_MAX_FETCH_SIZE_BYTES = float(os.getenv("SIIBRA_MAX_FETCH_SIZE_BYTES", 0.2 * 1024 ** 3))
SIIBRA_NG_VOL_USE_CACHE = eval(os.getenv("SIIBRA_NG_VOL_USE_CACHE", "False"))

with open(os.path.join(ROOT_DIR, "VERSION"), "r") as fp:
    __version__ = fp.read().strip()


@dataclass
class CompareMapsResult:
    intersection_over_union: float
    intersection_over_first: float
    intersection_over_second: float
    correlation: float
    weighted_mean_of_first: float
    weighted_mean_of_second: float


class TypePublication(TypedDict):
    citation: str
    url: str


T = TypeVar("T")


class InstanceTable(Generic[T], Iterable):
    """
    Lookup table for instances of a given class by name/id.
    Provide attribute-access and iteration to a set of named elements,
    given by a dictionary with keys of 'str' type.
    """

    def __init__(self, matchfunc=lambda a, b: a == b, elements=None):
        """
        Build an object lookup table from a dictionary with string keys, for easy
        attribute-like access, name autocompletion, and iteration.
        Matchfunc can be provided to enable inexact matching inside the index operator.
        It is a binary function, taking as first argument a value of the dictionary
        (ie. an object that you put into this glossary), and as second argument
        the index/specification that should match one of the objects, and returning a boolean.
        """

        assert hasattr(matchfunc, "__call__")
        if elements is None:
            self._elements: Dict[str, T] = {}
        else:
            assert isinstance(elements, dict)
            assert all(isinstance(k, str) for k in elements.keys())
            self._elements: Dict[str, T] = elements
        self._matchfunc = matchfunc
        self._dataframe_cached = None

    def add(self, key: str, value: T) -> None:
        """
        Add a key/value pair to the registry.

        Parameters
        ----------
            key (string): Unique name or key of the object
            value (object): The registered object
        """
        if key in self._elements:
            logger.error(
                f"Key {key} already in {__class__.__name__}, existing value will be replaced."
            )
        self._elements[key] = value

    def __dir__(self) -> Iterable[str]:
        """List of all object keys in the registry"""
        if isinstance(self[0], type):
            return list(self._elements.keys())
        else:
            return ["dataframe"] + list(self._elements.keys())

    def __str__(self) -> str:
        if len(self) > 0:
            return f"{self.__class__.__name__}:\n - " + "\n - ".join(self._elements.keys())
        else:
            return f"Empty {self.__class__.__name__}"

    def __repr__(self):
        return f"<{self.__class__.__name__} of {self[0].__class__}>"

    def __iter__(self) -> Iterator[T]:
        """Iterate over all objects in the registry"""
        return (w for w in self._elements.values())

    def __contains__(self, key: Union[str, T]) -> bool:
        """Test whether the given key or element is defined by the registry."""
        if isinstance(key, str):
            return key in self._elements
        return key in [item for _, item in self._elements.values()]

    def __len__(self) -> int:
        """Return the number of elements in the registry"""
        return len(self._elements)

    def __getitem__(self, spec) -> T:
        return self.get(spec)

    def get(self, spec) -> T:
        """Give access to objects in the registry by sequential index,
        exact key, or keyword matching. If the keywords match multiple objects,
        the first in sorted order is returned. If the specification does not match,
        a RuntimeError is raised.

        Parameters
        ----------
        spec: int, str
            Index or string specification of an object

        Returns
        -------
            Matched object
        """
        if spec is None:
            return None
        elif spec == "":
            raise IndexError(f"{__class__.__name__} indexed with empty string")
        matches = self.find(spec)
        if len(matches) == 0:
            raise IndexError(
                f"{__class__.__name__} has no entry matching the specification '{spec}'."
                f"Possible values are:\n" + str(self)
            )
        elif len(matches) == 1:
            return matches[0]
        else:
            try:
                S = sorted(matches, reverse=True)
            except TypeError:
                # not all object types support sorting, accept this
                S = matches
            largest = S[0]
            logger.info(
                f"Multiple elements matched the specification '{spec}' - the first in order was chosen: {largest}"
            )
            return largest

    def __sub__(self, obj) -> "InstanceTable[T]":
        """
        remove an object from the registry
        """
        if obj in self._elements.values():
            return InstanceTable[T](
                self._matchfunc, {k: v for k, v in self._elements.items() if v != obj}
            )
        else:
            return self

    def provides(self, spec) -> bool:
        """
        Returns True if an element that matches the given specification can be found
        (using find(), thus going beyond the matching of names only as __contains__ does)
        """
        matches = self.find(spec)
        return len(matches) > 0

    def find(self, spec) -> List[T]:
        """
        Return a list of items matching the given specification,
        which could be either the name or a specification that
        works with the matchfunc of the Glossary.
        """
        if isinstance(spec, str) and (spec in self._elements):
            return [self._elements[spec]]
        elif isinstance(spec, int) and (spec < len(self._elements)):
            return [list(self._elements.values())[spec]]
        else:
            # string matching on values
            matches = [v for v in self._elements.values() if self._matchfunc(v, spec)]
            if len(matches) == 0:
                # string matching on keys
                matches = [
                    self._elements[k]
                    for k in self._elements.keys()
                    if all(w.lower() in k.lower() for w in spec.split())
                ]
            return matches

    def values(self):
        return self._elements.values()

    def __getattr__(self, index) -> T:
        """Access elements by using their keys as attributes.
        Keys are auto-generated from the provided names to be uppercase,
        with words delimited using underscores.
        """
        if index in ["keys", "names"]:
            return list(self._elements.keys())
        elif index in self._elements:
            return self._elements[index]
        else:
            hint = ""
            if isinstance(index, str):
                import difflib

                closest = difflib.get_close_matches(
                    index, list(self._elements.keys()), n=3
                )
                if len(closest) > 0:
                    hint = f"Did you mean {' or '.join(closest)}?"
            raise AttributeError(f"Term '{index}' not in {__class__.__name__}. " + hint)

    @property
    def dataframe(self):
        if self._dataframe_cached is None:
            values = self._elements.values()
            attrs = []
            for i, val in enumerate(values):
                attrs.append({'name': val.name, 'species': str(val.species)})
                if hasattr(val, 'maptype'):
                    attrs[i].update(
                        {
                            attribute: val.__getattribute__(attribute).name
                            for attribute in ['parcellation', 'space', 'maptype']
                        }
                    )
            self._dataframe_cached = pd.DataFrame(index=list(self._elements.keys()), data=attrs)
        return self._dataframe_cached


class LoggingContext:
    def __init__(self, level):
        self.level = level

    def __enter__(self):
        self.old_level = logger.level
        logger.setLevel(self.level)

    def __exit__(self, et, ev, tb):
        logger.setLevel(self.old_level)


def set_log_level(level):
    logger.setLevel(level)


set_log_level(SIIBRA_LOG_LEVEL)
QUIET = LoggingContext("ERROR")
VERBOSE = LoggingContext("DEBUG")


def siibra_tqdm(iterable: Iterable[T] = None, *args, **kwargs):
    return tqdm(
        iterable,
        *args,
        disable=kwargs.pop("disable", False) or (logger.level > logging.INFO),
        **kwargs
    )


def create_key(name: str):
    """
    Creates an uppercase identifier string that includes only alphanumeric
    characters and underscore from a natural language name.
    """
    return re.sub(
        r" +",
        "_",
        "".join([e if e.isalnum() else " " for e in name]).upper().strip(),
    )


class MapIndex:
    """
    Identifies a unique region in a ParcellationMap, combining its labelindex (the "color") and mapindex (the number of the 3Dd map, in case multiple are provided).
    """

    def __init__(self, volume: int = None, label: int = None, fragment: str = None):
        if volume is None and label is None:
            raise ValueError(
                "At least volume or label need to be specified to build a valid map index."
            )
        if volume is not None:
            assert isinstance(volume, int)
        if label is not None:
            assert isinstance(label, int)
        self.volume = volume
        self.label = label
        self.fragment = fragment

    @classmethod
    def from_dict(cls, spec: dict):
        assert all(k in spec for k in ['volume', 'label'])
        return cls(
            volume=spec['volume'],
            label=spec['label'],
            fragment=spec.get('fragment')
        )

    def __str__(self):
        return f"(volume:{self.volume}, label:{self.label}, fragment:{self.fragment})"

    def __repr__(self):
        frag = f"'{self.fragment}'" if self.fragment else self.fragment
        return f"<{self.__class__.__name__}(volume={self.volume}, label={self.label}, fragment={frag})>"

    def __eq__(self, other):
        assert isinstance(other, self.__class__), f'Cannot compare {self.__class__} and {other.__class__}'
        return all([
            self.volume == other.volume,
            self.label == other.label,
            self.fragment == other.fragment
        ])

    def __hash__(self):
        return hash((self.volume, self.label, self.fragment))


class MapType(Enum):
    LABELLED = 1
    STATISTICAL = 2


REMOVE_FROM_NAME = [
    "hemisphere",
    " -",
    "-brain",
    "both",
    "Both",
]

REPLACE_IN_NAME = {
    "ctx-lh-": "left ",
    "ctx-rh-": "right ",
}


def clear_name(name):
    """ clean up a region name to the for matching"""
    result = name
    for word in REMOVE_FROM_NAME:
        result = result.replace(word, "")
    for search, repl in REPLACE_IN_NAME.items():
        result = result.replace(search, repl)
    return " ".join(w for w in result.split(" ") if len(w))


def snake2camel(s: str):
    """Converts a string in snake_case into CamelCase.
    For example: JULICH_BRAIN -> JulichBrain"""
    return "".join([w[0].upper() + w[1:].lower() for w in s.split("_")])


# getting nonzero pixels of pmaps is one of the most time consuming tasks when computing metrics,
# so we cache the nonzero coordinates of array objects at runtime.
NZCACHE = {}


def nonzero_coordinates(arr):
    # TODO: fix caching
    return np.c_[np.nonzero(arr > 0)]


def affine_scaling(affine):
    """Estimate approximate isotropic scaling factor of an affine matrix. """
    orig = np.dot(affine, [0, 0, 0, 1])
    unit_lengths = []
    for vec in np.identity(3):
        vec_phys = np.dot(affine, np.r_[vec, 1])
        unit_lengths.append(np.linalg.norm(orig - vec_phys))
    return np.prod(unit_lengths)


def compare_arrays(arr1: np.ndarray, affine1: np.ndarray, arr2: np.ndarray, affine2: np.ndarray):
    """
    Compare two arrays in physical space as defined by the given affine matrices.
    Matrices map voxel coordinates to physical coordinates.
    This function uses the object id to cache extraction of the nonzero coordinates.
    Repeated calls involving the same map will therefore be much faster as they
    will only access the image array if overlapping pixels are detected.

    It is recommended to install the indexed-gzip package,
    which will further speed this up.
    """
    a1, a2 = arr1.squeeze(), arr2.squeeze()

    def homog(XYZ):
        return np.c_[XYZ, np.ones(XYZ.shape[0])]

    def colsplit(XYZ):
        return np.split(XYZ, 3, axis=1)

    # Compute the nonzero voxels in map2 and their correspondences in map1
    XYZnz2 = nonzero_coordinates(a2)
    N2 = XYZnz2.shape[0]
    warp2on1 = np.dot(np.linalg.inv(affine1), affine2)
    XYZnz2on1 = (np.dot(warp2on1, homog(XYZnz2).T).T[:, :3] + 0.5).astype("int")

    # valid voxel pairs
    valid = np.all(
        np.logical_and.reduce(
            [
                XYZnz2on1 >= 0,
                XYZnz2on1 < arr1.shape[:3],
                XYZnz2 >= 0,
                XYZnz2 < arr2.shape[:3],
            ]
        ),
        1,
    )
    X1, Y1, Z1 = colsplit(XYZnz2on1[valid, :])
    X2, Y2, Z2 = colsplit(XYZnz2[valid, :])

    # intersection
    v1, v2 = a1[X1, Y1, Z1].squeeze(), a2[X2, Y2, Z2].squeeze()
    m1, m2 = ((_ > 0).astype("uint8") for _ in [v1, v2])
    intersection = np.minimum(m1, m2).sum()
    if intersection == 0:
        return CompareMapsResult(
            intersection_over_union=0,
            intersection_over_first=0,
            intersection_over_second=0,
            correlation=0,
            weighted_mean_of_first=0,
            weighted_mean_of_second=0,
        )

    # Compute the nonzero voxels in map1 with their correspondences in map2
    XYZnz1 = nonzero_coordinates(a1)
    N1 = XYZnz1.shape[0]
    warp1on2 = np.dot(np.linalg.inv(affine2), affine1)

    # Voxels referring to the union of the nonzero pixels in both maps
    XYZa1 = np.unique(np.concatenate((XYZnz1, XYZnz2on1)), axis=0)
    XYZa2 = (np.dot(warp1on2, homog(XYZa1).T).T[:, :3] + 0.5).astype("int")
    valid = np.all(
        np.logical_and.reduce(
            [XYZa1 >= 0, XYZa1 < arr1.shape[:3], XYZa2 >= 0, XYZa2 < arr2.shape[:3]]
        ),
        1,
    )
    Xa1, Ya1, Za1 = colsplit(XYZa1[valid, :])
    Xa2, Ya2, Za2 = colsplit(XYZa2[valid, :])

    # pearson's r wrt to full size image
    x = a1[Xa1, Ya1, Za1].squeeze()
    y = a2[Xa2, Ya2, Za2].squeeze()
    mu_x = x.sum() / a1.size
    mu_y = y.sum() / a2.size
    x0 = x - mu_x
    y0 = y - mu_y
    dem = np.sqrt(np.sum(x0 ** 2) * np.sum(y0 ** 2))
    if dem == 0:
        r = 0
    else:
        r = np.sum(np.multiply(x0, y0)) / dem

    bx = (x > 0).astype("uint8")
    by = (y > 0).astype("uint8")
    return CompareMapsResult(
        intersection_over_union=intersection / np.maximum(bx, by).sum(),
        intersection_over_first=intersection / N1,
        intersection_over_second=intersection / N2,
        correlation=r,
        weighted_mean_of_first=np.sum(x * y) / np.sum(y),
        weighted_mean_of_second=np.sum(x * y) / np.sum(x),
    )


def resample_img_to_img(
    source_img: Nifti1Image,
    target_img: Nifti1Image,
    interpolation: str = ""
) -> Nifti1Image:
    """
    Resamples to source image to match the target image according to target's
    affine. (A wrapper of `nilearn.image.resample_to_img`.)

    Parameters
    ----------
    source_img : Nifti1Image
    target_img : Nifti1Image
    interpolation : str, Default: "nearest" if the source image is a mask otherwise "linear".
        Can be 'continuous', 'linear', or 'nearest'. Indicates the resample method.

    Returns
    -------
    Nifti1Image
    """
    from nilearn._version import version as nilearn_version
    from packaging.version import Version

    interpolation = "nearest" if np.array_equal(np.unique(source_img.dataobj), [0, 1]) else "linear"
    kwargs = dict(
        source_img=source_img,
        target_img=target_img,
        interpolation=interpolation,
        force_resample=True,  # False is intended for testing. see nilearn docs
    )
    if Version(nilearn_version) >= Version("0.11.0"):
        # because nilearn>=0.11.0 don't support "copy_header" and python <= 3.8
        kwargs["copy_header"] = True  # use new default in nilearn >= 0.11.0

    resampled_img = resample_to_img(**kwargs)
    return resampled_img


def connected_components(
    imgdata: np.ndarray,
    background: int = 0,
    connectivity: int = 2,
    threshold: float = 0.0,
) -> Generator[Tuple[int, np.ndarray], None, None]:
    """
    Provide an iterator over connected components in the array. If the image
    data is float (such as probability maps), it will convert to a mask and
    then find the connected components.

    Note
    ----
    `Uses skimage.measure.label()` to determine foreground components.

    Parameters
    ----------
    imgdata : np.ndarray
    background_value : int, Default: 0
    connectivity : int, Default: 2
    threshold: float, Default: 0.0
        The threshold used to create mask from probability maps, i.e, anything
        below set to 0 and rest to 1.

    Yields
    ------
    Generator[Tuple[int, np.ndarray], None, None]
        tuple of integer label of the component and component as an ndarray in
        the shape of the original image.
    """
    from skimage import measure

    mask = (imgdata > threshold).astype('uint8')
    components = measure.label(mask, connectivity=connectivity, background=background)
    component_labels = np.unique(components)
    return (
        (label, (components == label).astype('uint8'))
        for label in component_labels
        if label > 0
    )


def unify_stringlist(L: list):
    """Adds asterisks to strings that appear multiple times, so the resulting
    list has only unique strings but still the same length, order, and meaning.
    For example:
        unify_stringlist(['a','a','b','a','c']) -> ['a','a*','b','a**','c']
    """
    assert all([isinstance(_, str) for _ in L])
    return [L[i] + "*" * L[:i].count(L[i]) for i in range(len(L))]


def create_gaussian_kernel(sigma=1, sigma_point=3):
    """
    Compute a 3D Gaussian kernel of the given bandwidth.
    """
    r = int(sigma_point * sigma)
    k_size = 2 * r + 1
    impulse = np.zeros((k_size, k_size, k_size))
    impulse[r, r, r] = 1
    kernel = gaussian(impulse, sigma)
    kernel /= kernel.sum()
    return kernel


def argmax_dim4(img, dim=-1):
    """
    Given a nifti image object with four dimensions, returns a modified object
    with 3 dimensions that is obtained by taking the argmax along one of the
    four dimensions (default: the last one). To distinguish the pure background
    voxels from the foreground voxels of channel 0, the argmax indices are
    incremented by 1 and label index 0 is kept to represent the background.
    """
    assert len(img.shape) == 4
    assert dim >= -1 and dim < 4
    newarr = np.asarray(img.dataobj).argmax(dim) + 1
    # reset the true background voxels to zero
    newarr[np.asarray(img.dataobj).max(dim) == 0] = 0
    return Nifti1Image(dataobj=newarr, header=img.header, affine=img.affine)


def MI(arr1, arr2, nbins=100, normalized=True):
    """
    Compute the mutual information between two 3D arrays, which need to have the same shape.

    Parameters
    ----------
    arr1: np.ndarray
        First 3D array
    arr2: np.ndarray
        Second 3D array
    nbins: int
        number of bins to use for computing the joint histogram (applies to intensity range)
    normalized: Boolean. Default: True
        if True, the normalized MI of arrays X and Y will be returned,
        leading to a range of values between 0 and 1. Normalization is
        achieved by NMI = 2*MI(X,Y) / (H(X) + H(Y)), where  H(x) is the entropy of X
    """

    assert all(len(arr.shape) == 3 for arr in [arr1, arr2])
    assert (all(arr.size > 0) for arr in [arr1, arr2])

    # compute the normalized joint 2D histogram as an
    # empirical measure of the joint probability of arr1 and arr2
    pxy, _, _ = np.histogram2d(arr1.ravel(), arr2.ravel(), bins=nbins)
    pxy /= pxy.sum()

    # extract the empirical probabilities of intensities
    # from the joint histogram
    px = np.sum(pxy, axis=1)  # marginal for x over y
    py = np.sum(pxy, axis=0)  # marginal for y over x

    # compute the mutual information
    px_py = px[:, None] * py[None, :]
    nzs = pxy > 0  # nonzero value indices
    MI = np.sum(pxy[nzs] * np.log(pxy[nzs] / px_py[nzs]))
    if not normalized:
        return MI

    # normalize, using the sum of their individual entropies H
    def entropy(p):
        nz = p > 0
        assert np.count_nonzero(nz) > 0
        return -np.sum(p[nz] * np.log(p[nz]))

    Hx, Hy = [entropy(p) for p in [px, py]]
    assert (Hx + Hy) > 0
    NMI = 2 * MI / (Hx + Hy)
    return NMI


def is_mesh(structure: Union[list, dict]):
    if isinstance(structure, dict):
        return all(k in structure for k in ["verts", "faces"])
    elif isinstance(structure, list):
        return all(map(is_mesh, structure))
    else:
        return False


def merge_meshes(meshes: list, labels: list = None):
    # merge a list of meshes into one
    # if meshes have no labels, a list of labels of the
    # same length as the number of meshes can
    # be supplied to add a labeling per sub mesh.

    assert len(meshes) > 0
    if len(meshes) == 1:
        return meshes[0]

    assert all('verts' in m for m in meshes)
    assert all('faces' in m for m in meshes)
    has_labels = all('labels' in m for m in meshes)
    if has_labels:
        assert labels is None

    nverts = np.cumsum([0] + [m['verts'].shape[0] for m in meshes[:-1]])
    verts = np.concatenate([m['verts'] for m in meshes])
    faces = np.concatenate([m['faces'] + N for m, N in zip(meshes, nverts)])
    if has_labels:
        labels = np.array([_ for m in meshes for _ in m['labels']])
        return {'verts': verts, 'faces': faces, 'labels': labels}
    elif labels is not None:
        assert len(labels) == len(meshes)
        labels = np.array(
            [labels[i] for i, m in enumerate(meshes) for v in m['verts']]
        )
        return {'verts': verts, 'faces': faces, 'labels': labels}
    else:
        return {'verts': verts, 'faces': faces}


class Species(Enum):

    HOMO_SAPIENS = 1
    RATTUS_NORVEGICUS = 2
    MUS_MUSCULUS = 3
    MACACA_FASCICULARIS = 4
    MACACA_MULATTA = 5
    MACACA_FUSCATA = 6
    CHLOROCEBUS_AETHIOPS_SABAEUS = 7
    CALLITHRIX_JACCHUS = 8

    UNSPECIFIED_SPECIES = 999

    @classmethod
    def decode(cls, spec: Union[str, 'Species', dict], fail_if_not_successful=True):

        MINDS_IDS = {
            "0ea4e6ba-2681-4f7d-9fa9-49b915caaac9": 1,
            "f3490d7f-8f7f-4b40-b238-963dcac84412": 2,
            "cfc1656c-67d1-4d2c-a17e-efd7ce0df88c": 3,
            "c541401b-69f4-4809-b6eb-82594fc90551": 4,
            "745712aa-fad1-47c4-8ab6-088063f78f64": 5,
            "ed8254b1-519c-4356-b1c9-7ead5aa1e3e1": 6,
            "e578d886-c55d-4174-976b-3cf43b142203": 7
        }

        OPENMINDS_IDS = {
            "97c070c6-8e1f-4ee8-9d28-18c7945921dd": 1,
            "ab532423-1fd7-4255-8c6f-f99dc6df814f": 2,
            "d9875ebd-260e-4337-a637-b62fed4aa91d": 3,
            "0b6df2b3-5297-40cf-adde-9443d3d8214a": 4,
            "3ad33ec1-5152-497d-9352-1cf4497e0edd": 5,
            "2ab3ecf5-76cc-46fa-98ab-309e3fd50f57": 6,
            "b8bf99e7-0914-4b65-a386-d785249725f1": 7
        }

        if isinstance(spec, Species):
            return spec
        elif isinstance(spec, str):
            # split it in case it is an actual uuid from KG
            if spec.split('/')[-1] in MINDS_IDS:
                return cls(MINDS_IDS[spec])
            if spec.split('/')[-1] in OPENMINDS_IDS:
                return cls(OPENMINDS_IDS[spec])
            key = cls.name_to_key(spec)
            if key in cls.__members__.keys():
                return getattr(cls, key)
        else:
            if isinstance(spec, (list, set)):
                next_specs = spec
            elif isinstance(spec, dict):
                next_specs = spec.values()
            else:
                raise ValueError(f"Species specification cannot be decoded: {spec}")
            for s in next_specs:
                result = cls.decode(s, fail_if_not_successful=False)
                if result is not None:
                    return result

        # if we get here, spec was not decoded into a species
        if fail_if_not_successful:
            raise ValueError(f"Species specification cannot be decoded: {spec}")
        else:
            return None

    @staticmethod
    def name_to_key(name: str):
        return re.sub(r'\s+', '_', name.strip()).upper()

    @staticmethod
    def key_to_name(key: str):
        return re.sub(r'_', ' ', key.strip()).lower()

    def __str__(self):
        return f"{self.name.lower().replace('_', ' ')}".capitalize()

    def __repr__(self):
        return f"{self.__class__.__name__}: {str(self)}"


def generate_uuid(string: str):
    if isinstance(string, str):
        b = string.encode("UTF-8")
    elif isinstance(string, Nifti1Image):
        b = string.to_bytes()
    else:
        raise ValueError(f"Cannot build uuid for parameter type {type(string)}")
    hex_string = md5(b).hexdigest()
    return str(UUID(hex=hex_string))


def translation_matrix(tx: float, ty: float, tz: float):
    """Construct a 3D homoegneous translation matrix."""
    return np.array([
        [1, 0, 0, tx],
        [0, 1, 0, ty],
        [0, 0, 1, tz],
        [0, 0, 0, 1]
    ])


def y_rotation_matrix(alpha: float):
    """Construct a 3D y axis rotation matrix."""
    return np.array([
        [math.cos(alpha), 0, math.sin(alpha), 0],
        [0, 1, 0, 0],
        [-math.sin(alpha), 0, math.cos(alpha), 0],
        [0, 0, 0, 1]
    ])



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import struct
from functools import wraps
from typing import Callable

cipher = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_-"
separator = "."
neg = "~"


def encode_number(n, float_flag=False):
    if float_flag:
        b = struct.pack("f", n)
        new_n = struct.unpack("i", b)
        return encode_int(new_n[0])
    else:
        return encode_int(n)


def encode_int(n):
    if not isinstance(n, int):
        raise ValueError("Cannot encode int")

    residual = None
    result = ""
    if n < 0:
        result += neg
        residual = n * -1
    else:
        residual = n

    while True:
        result = cipher[residual % 64] + result
        residual = math.floor(residual / 64)

        if residual == 0:
            break
    return result


def decode_int(n):
    neg_flag = False
    if n[-1] == neg:
        neg_flag = True
        n = n[:-1]

    result = 0
    for char in n:
        val = cipher.index(char)
        result = result * 64 + val

    if neg_flag:
        result = result * -1
    return result


def decode_number(n, float_flag=False):
    if float_flag:
        raise NotImplementedError
    return decode_int(n)


def post_process(post_process: Callable):
    def outer(fn):
        @wraps(fn)
        def inner(*args, **kwargs):
            val = fn(*args, **kwargs)
            return post_process(val, *args, **kwargs)

        return inner

    return outer



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Metadata connection to EBRAINS datasets."""

import re
from abc import ABC, abstractmethod
from hashlib import md5
from typing import Union, List
try:
    from typing import TypedDict
except ImportError:
    # support python 3.7
    from typing_extensions import TypedDict

from .requests import MultiSourcedRequest, GitlabProxy, GitlabProxyEnum


class EbrainsDatasetUrl(TypedDict):
    url: str


EbrainsDatasetPerson = TypedDict("EbrainsDatasetPerson", {
    "@id": str,
    "schema.org/shortName": str,
    "identifier": str,
    "shortName": str,
    "name": str,
})

EbrainsDatasetEmbargoStatus = TypedDict("EbrainsDatasetEmbargoStatus", {
    "@id": str,
    "name": str,
    "identifier": List[str]
})


class EbrainsBaseDataset(ABC):
    @property
    @abstractmethod
    def id(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def urls(self) -> List[EbrainsDatasetUrl]:
        raise NotImplementedError

    @property
    @abstractmethod
    def description(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def contributors(self) -> List[EbrainsDatasetPerson]:
        raise NotImplementedError

    @property
    @abstractmethod
    def ebrains_page(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def custodians(self) -> List[EbrainsDatasetPerson]:
        raise NotImplementedError

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, o: object) -> bool:
        return hasattr(o, "id") and self.id == o.id

    def match(self, spec: Union[str, 'EbrainsBaseDataset']) -> bool:
        """
        Checks if the given specification describes this dataset.

        Parameters
        ----------
        spec (str, EbrainsBaseDataset)
            specification to be matched.
        Returns
        -------
        bool
        """
        if spec is self:
            return True
        if isinstance(spec, str):
            return self.id == spec
        raise RuntimeError(
            f"Cannot match {spec.__class__}, must be either str or EbrainsBaseDataset"
        )

    @property
    def LICENSE(self) -> str:
        license_ = self._detail.get("license", [])
        if len(license_) > 0:
            return license_ if isinstance(license_, str) else '\n'.join(license_)
        return None


class EbrainsDataset(EbrainsBaseDataset):
    """Ebrains dataset v1 connection"""

    def __init__(self, id, name=None, embargo_status: List[EbrainsDatasetEmbargoStatus] = None, *, cached_data=None):
        super().__init__()

        self._id = id
        self._name = name
        self._cached_data = cached_data
        self.embargo_status = embargo_status

        if id is None:
            raise TypeError("Dataset id is required")

        match = re.search(r"([a-f0-9-]+)$", id)
        if not match:
            raise ValueError(
                f"{self.__class__.__name__} initialized with invalid id: {self.id}"
            )

    @property
    def id(self) -> str:
        return self._id

    @property
    def _detail(self):
        if not self._cached_data:
            match = re.search(r"([a-f0-9-]+)$", self.id)
            instance_id = match.group(1)
            self._cached_data = MultiSourcedRequest(
                requests=[
                    GitlabProxy(
                        GitlabProxyEnum.DATASET_V1,
                        instance_id=instance_id,
                    ),
                ]
            ).data
        return self._cached_data

    @property
    def name(self) -> str:
        if self._name is None:
            self._name = self._detail.get("name")
        return self._name

    @property
    def urls(self) -> List[EbrainsDatasetUrl]:
        return [
            {
                "url": f if f.startswith("http") else f"https://doi.org/{f}",
            }
            for f in self._detail.get("kgReference", [])
        ]

    @property
    def description(self) -> str:
        return self._detail.get("description")

    @property
    def contributors(self) -> List[EbrainsDatasetPerson]:
        return self._detail.get("contributors")

    @property
    def ebrains_page(self):
        return f"https://search.kg.ebrains.eu/instances/{self.id}"

    @property
    def custodians(self) -> EbrainsDatasetPerson:
        return self._detail.get("custodians")


class EbrainsV3DatasetVersion(EbrainsBaseDataset):
    @staticmethod
    def _parse_person(d: dict) -> EbrainsDatasetPerson:
        assert "https://openminds.ebrains.eu/core/Person" in d.get("type"), "Cannot convert a non person to a person dict!"
        _id = d.get("id")
        name = f"{d.get('givenName')} {d.get('familyName')}"
        return {
            '@id': _id,
            'schema.org/shortName': name,
            'identifier': _id,
            'shortName': name,
            'name': name
        }

    def __init__(self, id, *, cached_data=None) -> None:
        super().__init__()

        self._id = id
        self._cached_data = cached_data

    @property
    def _detail(self):
        if not self._cached_data:
            match = re.search(r"([a-f0-9-]+)$", self._id)
            instance_id = match.group(1)
            self._cached_data = MultiSourcedRequest(
                requests=[
                    GitlabProxy(
                        GitlabProxyEnum.DATASETVERSION_V3,
                        instance_id=instance_id,
                    ),
                ]
            ).data
        return self._cached_data

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        fullname = self._detail.get("fullName")
        if not fullname:
            for dataset in self.is_version_of:
                if fullname:
                    break
                fullname = dataset.name
        version_id = self._detail.get("versionIdentifier")
        return f"{fullname} ({version_id})"

    @property
    def is_version_of(self):
        if not hasattr(self, "_is_version_of"):
            self._is_version_of = [EbrainsV3Dataset(id=id.get("id")) for id in self._detail.get("isVersionOf", [])]
        return self._is_version_of

    @property
    def urls(self) -> List[EbrainsDatasetUrl]:
        return [{
            "url": doi.get("identifier", None)
        } for doi in self._detail.get("doi", [])]

    @property
    def description(self) -> str:
        description = self._detail.get("description")
        for ds in self.is_version_of:
            if description:
                break
            description = ds.description
        return description or ""

    @property
    def contributors(self) -> List[EbrainsDatasetPerson]:
        return [EbrainsV3DatasetVersion._parse_person(d) for d in self._detail.get("author", [])]

    @property
    def ebrains_page(self) -> str:
        if len(self.urls) > 0:
            return self.urls[0].get("url")
        return None

    @property
    def custodians(self) -> EbrainsDatasetPerson:
        return [EbrainsV3DatasetVersion._parse_person(d) for d in self._detail.get("custodian", [])]

    @property
    def version_changelog(self):
        return self._detail.get("versionInnovation", "")

    @property
    def version_identifier(self):
        return self._detail.get("versionIdentifier", "")


class EbrainsV3Dataset(EbrainsBaseDataset):
    def __init__(self, id, *, cached_data=None) -> None:
        super().__init__()

        self._id = id
        self._cached_data = cached_data
        self._contributers = None

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._detail.get("fullName")

    @property
    def urls(self) -> List[EbrainsDatasetUrl]:
        return [{
            "url": doi.get("identifier", None)
        } for doi in self._detail.get("doi", [])]

    @property
    def _detail(self):
        if not self._cached_data:
            match = re.search(r"([a-f0-9-]+)$", self._id)
            instance_id = match.group(1)
            self._cached_data = MultiSourcedRequest(
                requests=[
                    GitlabProxy(
                        GitlabProxyEnum.DATASET_V3,
                        instance_id=instance_id,
                    ),
                ]
            ).data
        return self._cached_data

    @property
    def description(self) -> str:
        return self._detail.get("description", "")

    @property
    def contributors(self):
        if self._contributers is None:
            contributors = {}
            for version_id in self.version_ids:
                contributors.update(
                    {c['@id']: c for c in EbrainsV3DatasetVersion(version_id).contributors}
                )
            self._contributers = list(contributors.values())
        return self._contributers

    @property
    def ebrains_page(self) -> str:
        if len(self.urls) > 0:
            return self.urls[0].get("url")
        return None

    @property
    def custodians(self) -> EbrainsDatasetPerson:
        return [EbrainsV3DatasetVersion._parse_person(d) for d in self._detail.get("custodian", [])]

    @property
    def version_ids(self) -> List['str']:
        return [version.get("id") for version in self._detail.get("versions", [])]


class GenericDataset:

    def __init__(
        self,
        name: str = None,
        contributors: List[str] = None,
        url: str = None,
        description: str = None,
        license: str = None
    ):
        self._name = name
        self._contributors = contributors
        self._url = url
        self._description = description
        self._license = license

    @property
    def contributors(self):
        return [{"name": cont} for cont in self._contributors]

    @property
    def id(self) -> str:
        return md5(self.name.encode('utf-8')).hexdigest()

    @property
    def name(self) -> str:
        return self._name

    @property
    def LICENSE(self) -> str:
        return self._license

    @property
    def urls(self) -> List[EbrainsDatasetUrl]:
        return [{"url": self._url}]

    @property
    def description(self) -> str:
        return self._description

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, o: object) -> bool:
        return hasattr(o, "id") and self.id == o.id
